# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import copy
from functools import total_ordering
from openc3.accessors.binary_accessor import BinaryAccessor


@total_ordering
class StructureItem:
    create_index = 0

    # Valid data types adds DERIVED to those defined by BinaryAccessor
    DATA_TYPES = BinaryAccessor.DATA_TYPES + ["DERIVED"]

    # Create a StructureItem by setting all the attributes. It
    # calls all the setter routines to do the attribute verification and then
    # verifies the overall integrity.
    #
    # self.param name [String] The item name
    # self.param bit_offset [Integer] Offset to the item starting at 0
    # self.param bit_size [Integer] Size of the items in bits
    # self.param data_type [Symbol] {DATA_TYPES}
    # self.param endianness [Symbol] {BinaryAccessor::ENDIANNESS}
    # self.param array_size [Integer, None] Size of the array item in bits. For
    #   example, if the bit_size is 8, an array_size of 16 holds two values.
    # self.param overflow [Symbol] {BinaryAccessor::OVERFLOW_TYPES}
    def __init__(
        self,
        name,
        bit_offset,
        bit_size,
        data_type,
        endianness,
        array_size=None,
        overflow="ERROR",
    ):
        self.structure_item_constructed = False
        # Assignment order matters due to verifications!
        self.name = name
        self.key = name  # Key defaults to name as given (not upcased)
        self.endianness = endianness
        self.data_type = data_type
        self.bit_offset = bit_offset
        self.original_bit_offset = bit_offset
        self.bit_size = bit_size
        self.original_bit_size = bit_size
        self.array_size = array_size
        self.original_array_size = array_size
        self.overflow = overflow
        self.overlap = False
        self.variable_bit_size = False
        self.create_index = StructureItem.create_index
        StructureItem.create_index += 1
        self.structure_item_constructed = True
        self.verify_overall()

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, name):
        if not isinstance(name, str):
            raise TypeError(f"name must be a String but is a {name.__class__.__name__}")
        if len(name) == 0:
            raise ValueError("name must contain at least one character")

        self.__name = name.upper()
        if self.structure_item_constructed:
            self.verify_overall()

    @property
    def key(self):
        return self.__key

    @key.setter
    def key(self, key):
        if not isinstance(key, str):
            raise TypeError(f"key must be a String but is a {key.__class__.__name__}")
        if len(key) == 0:
            raise ValueError("key must contain at least one character")
        self.__key = key

    @property
    def endianness(self):
        return self.__endianness

    @endianness.setter
    def endianness(self, endianness):
        if not isinstance(endianness, str):
            raise TypeError(f"{self.name}: endianness must be a String but is a {endianness.__class__.__name__}")
        if endianness not in BinaryAccessor.ENDIANNESS:
            raise ValueError(
                f"{self.name}: unknown endianness: {endianness} - Must be 'BIG_ENDIAN' or 'LITTLE_ENDIAN'"
            )
        self.__endianness = endianness
        if self.structure_item_constructed:
            self.verify_overall()

    @property
    def bit_offset(self):
        return self.__bit_offset

    @bit_offset.setter
    def bit_offset(self, bit_offset):
        if not isinstance(bit_offset, int):
            raise TypeError(f"{self.name}: bit_offset must be an Integer")

        byte_aligned = (bit_offset % 8) == 0
        if (self.data_type == "FLOAT" or self.data_type == "STRING" or self.data_type == "BLOCK") and not byte_aligned:
            raise ValueError(
                f"{self.name}: bit_offset for 'FLOAT', 'STRING', and 'BLOCK' items must be byte aligned"
            )

        if self.data_type == "DERIVED" and bit_offset != 0:
            raise ValueError(f"{self.name}: DERIVED items must have bit_offset of zero")

        self.__bit_offset = bit_offset
        if self.structure_item_constructed:
            self.verify_overall()

    @property
    def bit_size(self):
        return self.__bit_size

    @bit_size.setter
    def bit_size(self, bit_size):
        if not isinstance(bit_size, int):
            raise TypeError(f"{self.name}: bit_size must be an Integer")

        byte_multiple = (bit_size % 8) == 0
        if bit_size <= 0 and self.data_type == "FLOAT":
            raise ValueError(f"{self.name}: bit_size cannot be negative or zero for 'FLOAT' items: {bit_size}")
        if (self.data_type == "STRING" or self.data_type == "BLOCK") and not byte_multiple:
            raise ValueError(f"{self.name}: bit_size for STRING and BLOCK items must be byte multiples")
        if self.data_type == "FLOAT" and bit_size != 32 and bit_size != 64:
            raise ValueError(f"{self.name}: bit_size for FLOAT items must be 32 or 64. Given: {bit_size}")
        if self.data_type == "DERIVED" and bit_size != 0:
            raise ValueError(f"{self.name}: DERIVED items must have bit_size of zero")

        self.__bit_size = bit_size
        if self.structure_item_constructed:
            self.verify_overall()

    @property
    def data_type(self):
        return self.__data_type

    @data_type.setter
    def data_type(self, data_type):
        if not isinstance(data_type, str):
            raise TypeError(
                f"{self.name}: data_type must be a str but {data_type} is a {type(data_type).__name__}"
            )
        if data_type not in StructureItem.DATA_TYPES:
            raise ValueError(
                f"{self.name}: unknown data_type: {data_type} - Must be 'INT', 'UINT', 'FLOAT', 'STRING', 'BLOCK', or 'DERIVED'"
            )

        self.__data_type = data_type
        if self.structure_item_constructed:
            self.verify_overall()

    @property
    def array_size(self):
        return self.__array_size

    @array_size.setter
    def array_size(self, array_size):
        if array_size is not None:
            if not isinstance(array_size, int):
                raise TypeError(f"{self.name}: array_size must be an Integer")
            if not (self.bit_size == 0 or (array_size % self.bit_size == 0) or array_size < 0):
                raise ValueError(f"{self.name}: array_size must be a multiple of bit_size")
            if self.bit_size <= 0:
                raise ValueError(f"{self.name}: bit_size cannot be negative or zero for array items")

        self.__array_size = array_size
        if self.structure_item_constructed:
            self.verify_overall()

    @property
    def overflow(self):
        return self.__overflow

    @overflow.setter
    def overflow(self, overflow):
        if not isinstance(overflow, str):
            raise TypeError(f"{self.name}: overflow type must be a String")

        if overflow not in BinaryAccessor.OVERFLOW_TYPES:
            raise ValueError(
                f"{self.name}: unknown overflow type: {overflow} - Must be 'ERROR', 'ERROR_ALLOW_HEX', 'TRUNCATE', or 'SATURATE'"
            )

        self.__overflow = overflow
        if self.structure_item_constructed:
            self.verify_overall()

    @property
    def variable_bit_size(self):
        return self.__variable_bit_size

    @variable_bit_size.setter
    def variable_bit_size(self, variable_bit_size):
        if variable_bit_size:
            if not isinstance(variable_bit_size, dict):
                raise TypeError(f"{self.name}: variable_bit_size must be a dict")
            if not isinstance(variable_bit_size["length_item_name"], str):
                raise TypeError(f"{self.name}: variable_bit_size['length_item_name'] must be a String")
            if not isinstance(variable_bit_size["length_value_bit_offset"], int):
                raise ValueError(f"{self.name}: variable_bit_size['length_value_bit_offset'] must be an Integer")
            if not isinstance(variable_bit_size["length_bits_per_count"], int):
                raise ValueError(f"{self.name}: variable_bit_size['length_bits_per_count'] must be an Integer")
        self.__variable_bit_size = variable_bit_size
        if self.structure_item_constructed:
            self.verify_overall()

    def __eq__(self, other):
        # Sort by first bit_offset, then bit_size, then create_index
        if self.bit_offset == other.bit_offset:
            if self.bit_size == other.bit_size:
                if self.create_index:
                    return self.create_index == other.create_index
            else:
                return self.bit_size == other.bit_size
        else:
            return self.bit_offset == other.bit_offset

    def __lt__(self, other):
        if self.bit_offset == other.bit_offset:
            if self.bit_size == other.bit_size:
                if self.create_index:
                    return self.create_index < other.create_index
            else:
                return self.bit_size < other.bit_size
        else:
            if ((self.bit_offset >= 0) and (other.bit_offset >= 0)) or (
                (self.bit_offset < 0) and (other.bit_offset < 0)
            ):
                # Both Have Same Sign
                return self.bit_offset < other.bit_offset
            else:
                # Different signs
                return self.bit_offset > other.bit_offset

    # Make a light weight clone of this item
    def clone(self):
        item = copy.copy(self)
        # Since we're copying and not calling the constructor
        # we have to manually update the create_index
        item.create_index = StructureItem.create_index
        StructureItem.create_index += 1
        return item

    @classmethod
    def from_json(cls, structure):
        # Convert strings to symbols
        endianness = structure.get("endianness")
        data_type = structure.get("data_type")
        array_size = structure.get("array_size")
        overflow = structure.get("overflow")
        si = StructureItem(
            structure["name"],
            structure["bit_offset"],
            structure["bit_size"],
            data_type,
            endianness,
            array_size,
            overflow,
        )
        si.key = structure.get("key", structure["name"])
        return si

    def as_json(self):
        structure = {}
        structure["name"] = self.name
        structure["key"] = self.key
        structure["bit_offset"] = self.original_bit_offset
        structure["bit_size"] = self.original_bit_size
        structure["data_type"] = self.data_type
        structure["endianness"] = self.endianness
        structure["array_size"] = self.original_array_size
        structure["overflow"] = self.overflow
        return structure

    def little_endian_bit_field(self):
        if self.endianness != "LITTLE_ENDIAN":
            return False
        if not (self.data_type == "INT" or self.data_type == "UINT"):
            return False
        # If we're not byte aligned we're a bit field
        if not (self.bit_offset % 8) == 0:
            return True
        # If we don't have an even number of bytes we're a bit field
        if not self.even_byte_multiple():
            return True
        return False

    # Verifies overall integrity of the StructureItem by checking for correct
    # LITTLE_ENDIAN bit fields
    def verify_overall(self):
        # Verify negative bit_offset conditions
        if self.bit_offset < 0:
            if self.bit_size < 0:
                raise ValueError(
                    f"{self.name}: Can't define an item with negative bit_size {self.bit_size} and negative bit_offset {self.bit_offset}"
                )
            if self.array_size and self.array_size < 0:
                raise ValueError(
                    f"{self.name}: Can't define an item with negative array_size {self.array_size} and negative bit_offset {self.bit_offset}"
                )
            if self.array_size and self.array_size > abs(self.bit_offset):
                raise ValueError(
                    f"{self.name}: Can't define an item with array_size {self.array_size} greater than negative bit_offset {self.bit_offset}"
                )
            elif self.bit_size > abs(self.bit_offset):
                raise ValueError(
                    f"{self.name}: Can't define an item with bit_size {self.bit_size} greater than negative bit_offset {self.bit_offset}"
                )
        else:
            # Verify bounds on little-endian bit fields
            if self.little_endian_bit_field():
                # Bitoffset always refers to the most significant bit of a bitfield
                num_bytes = int(((self.bit_offset % 8) + self.bit_size - 1) / 8) + 1
                upper_bound = self.bit_offset / 8
                lower_bound = upper_bound - num_bytes + 1

                if lower_bound < 0:
                    raise ValueError(
                        f"{self.name}: LITTLE_ENDIAN bitfield with bit_offset {self.bit_offset} and bit_size {self.bit_size} is invalid"
                    )

    def even_byte_multiple(self):
        if self.bit_size in [8, 16, 32, 64]:
            return True
        else:
            return False



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.


from openc3.conversions.conversion import Conversion


# Performs a polynomial conversion on the value
class PolynomialConversion(Conversion):
    # Initializes the conversion with the given polynomial coefficients. Sets
    # the converted_type to :FLOAT and the converted_bit_size to 64.
    #
    # @param coeffs [Array<Float>] The polynomial coefficients
    def __init__(self, *coeffs):
        super().__init__()
        self.coeffs = [float(coeff) for coeff in coeffs]
        self.converted_type = "FLOAT"
        self.converted_bit_size = 64
        self.params = coeffs

    # @param (see Conversion#call)
    # @return [Float] The value with the polynomial applied
    def call(self, value, myself, buffer):
        value = float(value)

        # Handle C0
        result = self.coeffs[0]

        # Handle Coefficients raised to a power
        raised_to_power = 1.0
        for coeff in self.coeffs[1:]:
            raised_to_power *= value
            result += coeff * raised_to_power
        return result

    # @return [String] Class followed by the list of coefficients
    def __str__(self):
        result = ""
        for index in range(0, len(self.coeffs)):
            if index == 0:
                result += f"{self.coeffs[index]}"
            elif index == 1:
                result += f" + {self.coeffs[index]}x"
            else:
                result += f" + {self.coeffs[index]}x^{index}"
        return result

    # @param (see Conversion#to_config)
    # @return [String] Config fragment for this conversion
    def to_config(self, read_or_write):
        return f"    POLY_{read_or_write}_CONVERSION {' '.join(self.coeffs)}\n"



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import json
import openc3.script
from openc3.environment import OPENC3_SCOPE

def table_create_binary(definition: str, scope: str = OPENC3_SCOPE):
    data = {}
    data['definition'] = definition
    response = openc3.script.API_SERVER.request("post", "/openc3-api/tables/generate", data=data, json=True, scope=scope)
    return _handle_response(response, 'Failed to create binary')

def table_create_report(filename: str, definition: str, table_name: str = None, scope: str = OPENC3_SCOPE):
    data = {}
    data['binary'] = filename
    data['definition'] = definition
    if table_name:
        data['table_name'] = table_name
    response = openc3.script.API_SERVER.request("post", "/openc3-api/tables/report", data=data, json=True, scope=scope)
    return _handle_response(response, 'Failed to create binary')

# Helper method to handle the response
def _handle_response(response, error_message):
    if response is None:
        return None
    if response.status_code >= 400:
        result = json.loads(response.text)
        raise RuntimeError(f"{error_message} due to {result['message']}")
    # Not sure why the response body is empty (on delete) but check for that
    if response.text is None or len(response.text) == 0:
        return None
    else:
        return json.loads(response.text)



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.top_level import get_class_from_module
from openc3.utilities.string import filename_to_module, filename_to_class_name
from openc3.processors.processor import Processor


class ProcessorParser:
    # @param parser [ConfigParser] Configuration parser
    # @param packet [Packet] The current packet
    # @param cmd_or_tlm [String] Whether this is a command or telemetry packet
    @classmethod
    def parse(cls, parser, packet, cmd_or_tlm):
        parser = ProcessorParser(parser)
        parser.verify_parameters(cmd_or_tlm)
        parser.create_processor(packet)

    # @param parser [ConfigParser] Configuration parser
    def __init__(self, parser):
        self.parser = parser

    # @param cmd_or_tlm [String] Whether this is a command or telemetry packet
    def verify_parameters(self, cmd_or_tlm):
        if cmd_or_tlm.upper() == "COMMAND":
            raise self.parser.error("PROCESSOR only applies to telemetry packets")

        self.usage = "PROCESSOR <PROCESSOR NAME> <PROCESSOR CLASS FILENAME> <PROCESSOR SPECIFIC OPTIONS>"
        self.parser.verify_num_parameters(2, None, self.usage)

    # @param packet [Packet] The packet the processor should be added to
    def create_processor(self, packet):
        try:
            klass = get_class_from_module(
                filename_to_module(self.parser.parameters[1]),
                filename_to_class_name(self.parser.parameters[1]),
            )

            if len(self.parser.parameters) > 2:
                processor = klass(*self.parser.parameters[2 : (len(self.parser.parameters))])
            else:
                processor = klass()
            if not isinstance(processor, Processor):
                raise TypeError(f"processor must be a Processor but is a {processor.__class__.__name__}")

            processor.name = self._get_processor_name()
            packet.processors[processor.name] = processor
        except Exception as err:
            raise self.parser.error(err, self.usage)

    def _get_processor_name(self):
        return self.parser.parameters[0].upper()



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import queue
import select
import socket
import threading
import multiprocessing
from openc3.interfaces.stream_interface import StreamInterface
from openc3.streams.tcpip_socket_stream import TcpipSocketStream
from openc3.config.config_parser import ConfigParser
from openc3.utilities.logger import Logger
from openc3.top_level import kill_thread, close_socket


# Data class which stores the interface and associated information
class InterfaceInfo:
    # attr_reader :interface, :hostname, :host_ip, :port

    def __init__(self, interface, hostname, host_ip, port):
        self.interface = interface
        self.hostname = hostname
        self.host_ip = host_ip
        self.port = port


# TCP/IP Server which can both read and write on a single port or two
# independent ports. A listen thread is setup which waits for client
# connections. For each connection to the read port, a thread is spawned that
# calls the read method from the interface. This data is then
# available by calling the TcpipServer read method. For each connection to the
# write port, a thread is spawned that calls the write method from the
# interface when data is send to the TcpipServer via the write method.
class TcpipServerInterface(StreamInterface):
    # Callback method to call when a new client connects to the write port.
    # This method will be called with the Interface as the only argument.
    # attr_accessor :write_connection_callback
    # Callback method to call when a new client connects to the read port.
    # This method will be called with the Interface as the only argument.
    # attr_accessor :read_connection_callback
    # @return [StreamLogPair] StreamLogPair instance or nil
    # attr_accessor :stream_log_pair
    # @return [String] The ip address to bind to.  Default to ANY (0.0.0.0)
    # attr_accessor :listen_address

    # @param write_port [Integer] The server write port. Clients should connect
    #   and expect to receive data from this port.
    # @param read_port [Integer] The server read port. Clients should connect
    #   and expect to send data to this port.
    # @param write_timeout [Float] Seconds to wait before aborting writes
    # @param read_timeout [Float|nil] Seconds to wait before aborting reads.
    #   Pass nil to block until the read is complete.
    # @param protocol_type [String] The name of the stream to
    #   use for both the read and write ports. This name is combined with
    #   'Protocol' to result in a OpenC3 Protocol class.
    # @param protocol_args [Array] Arguments to pass to the Protocol
    def __init__(
        self,
        write_port,
        read_port,
        write_timeout,
        read_timeout,
        protocol_type=None,
        *protocol_args,
    ):
        super().__init__(protocol_type, protocol_args)
        self.write_port = ConfigParser.handle_none(write_port)
        if self.write_port is not None:
            self.write_port = int(self.write_port)
        self.read_port = ConfigParser.handle_none(read_port)
        if self.read_port is not None:
            self.read_port = int(self.read_port)
        self.write_timeout = ConfigParser.handle_none(write_timeout)
        if self.write_timeout is not None:
            self.write_timeout = float(self.write_timeout)
        self.read_timeout = ConfigParser.handle_none(read_timeout)
        if self.read_timeout is not None:
            self.read_timeout = float(self.read_timeout)
        self.listen_sockets = []
        self.listen_pipes = []
        self.listen_threads = []
        self.read_threads = []
        self.write_thread = None
        self.write_raw_thread = None
        self.write_interface_infos = []
        self.read_interface_infos = []
        self.write_queue = None
        if self.write_port:
            self.write_queue = queue.Queue()
        self.write_raw_queue = None
        if self.write_port:
            self.write_raw_queue = queue.Queue()
        self.read_queue = None
        if self.read_port:
            self.read_queue = queue.Queue()
        self.write_condition_variable = None
        if self.write_port:
            self.write_condition_variable = threading.Condition(self.write_mutex)
        self.write_raw_mutex = None
        if self.write_port:
            self.write_raw_mutex = threading.Lock()
        self.write_raw_condition_variable = None
        if self.write_port:
            self.write_raw_condition_variable = threading.Condition(self.write_raw_mutex)
        self.write_connection_callback = None
        self.read_connection_callback = None
        self.stream_log_pair = None
        self.raw_logging_enabled = False
        self.connection_mutex = threading.Lock()
        self.listen_address = "0.0.0.0"

        if not ConfigParser.handle_none(read_port):
            self.read_allowed = False
        if not ConfigParser.handle_none(write_port):
            self.write_allowed = False
        if not ConfigParser.handle_none(write_port):
            self.write_raw_allowed = False

        self._connected = False

    def connected(self):
        return self._connected

    def connection_string(self):
        if self.write_port == self.read_port:
            return f"listening on {self.listen_address}:{self.write_port} (R/W)"
        result = "listening on"
        if self.write_port:
            result += f" {self.listen_address}:{self.write_port} (write)"
        if self.read_port:
            result += f" {self.listen_address}:{self.read_port} (read)"
        return result

    # Create the read and write port listen threads. Incoming connections will
    # spawn separate threads to process the reads and writes.
    def connect(self):
        self.cancel_threads = False
        if self.read_queue:
            while not self.read_queue.empty():
                self.read_queue.get_nowait()
        if self.write_port == self.read_port:  # One socket
            self._start_listen_thread(self.read_port, True, True)
        else:
            if self.write_port:
                self._start_listen_thread(self.write_port, True, False)
            if self.read_port:
                self._start_listen_thread(self.read_port, False, True)

        if self.write_port:
            self.write_thread = threading.Thread(target=self._write_thread_body, daemon=True)
            self.write_thread.start()

            self.write_raw_thread = threading.Thread(target=self._write_raw_thread_body, daemon=True)
            self.write_raw_thread.start()
        else:
            self.write_thread = None
            self.write_raw_thread = None
        super().connect()
        self._connected = True

    # Shutdowns the listener threads for both the read and write ports as well
    # as any client connections.
    def disconnect(self):
        self.cancel_threads = True
        if self.read_queue:
            self.read_queue.put(None)
        for pipe in self.listen_pipes:
            pipe.send(".")
        self.listen_pipes = []

        # Shutdown listen thread(s)
        for listen_thread in self.listen_threads:
            kill_thread(self, listen_thread)
        self.listen_threads = []

        # Shutdown listen socket(s)
        for listen_socket in self.listen_sockets:
            close_socket(listen_socket)
        # Ok may have been closed by the thread
        self.listen_sockets = []

        # This will unblock read threads
        self._shutdown_interfaces(self.read_interface_infos)

        for thread in self.read_threads:
            kill_thread(self, thread)
        self.read_threads = []

        if self.write_thread:
            with self.write_condition_variable:
                self.write_condition_variable.notify_all()
            kill_thread(self, self.write_thread)
            self.write_thread = None
        if self.write_raw_thread:
            with self.write_raw_condition_variable:
                self.write_raw_condition_variable.notify_all()
            kill_thread(self, self.write_raw_thread)
            self.write_raw_thread = None

        self._shutdown_interfaces(self.write_interface_infos)
        self._connected = False
        super().disconnect()

    # Gracefully kill all the threads
    def graceful_kill(self):
        # This method is just here to prevent warnings
        pass

    # @return [Packet] Latest packet read from any of the connected clients.
    #   Note this method blocks until data is available.
    def read(self):
        if not self.connected():
            raise RuntimeError(f"Interface not connected for read: {self.name}")
        if not self.read_allowed:
            raise RuntimeError(f"Interface not readable: {self.name}")

        try:
            packet = self.read_queue.get(block=True)
        except queue.Empty:
            return None

        if packet is not None:
            self.read_count += 1
        return packet

    # @param packet [Packet] Packet to write to all clients connected to the
    #   write port.
    def write(self, packet):
        if not self.connected():
            raise RuntimeError(f"Interface not connected for write: {self.name}")
        if not self.write_allowed:
            raise RuntimeError(f"Interface not writeable: {self.name}")

        self.write_count += 1
        self.write_queue.put(packet.clone())
        try:
            self.write_condition_variable.notify_all()
        except RuntimeError as error:
            if "cannot notify on un-acquired lock" in repr(error):
                pass
            else:
                raise error

    # @param data [String] Data to write to all clients connected to the
    #   write port.
    def write_raw(self, data):
        if not self.connected():
            raise RuntimeError(f"Interface not connected for write_raw: {self.name}")
        if not self.write_raw_allowed:
            raise RuntimeError(f"Interface not write-rawable: {self.name}")

        self.write_raw_queue.put(data)
        try:
            self.write_raw_condition_variable.notify_all()
        except RuntimeError as error:
            if "cannot notify on un-acquired lock" in repr(error):
                return data
            else:
                raise error

    # @return [Integer] The number of packets waiting on the read queue
    def read_queue_size(self):
        if self.read_queue:
            return self.read_queue.qsize()
        else:
            return 0

    # @return [Integer] The number of packets waiting on the write queue
    def write_queue_size(self):
        if self.write_queue:
            return self.write_queue.qsize()
        else:
            return 0

    # @return [Integer] The number of connected clients
    def num_clients(self):
        interfaces = []
        for wii in self.write_interface_infos:
            interfaces.append(wii.interface)
        for rii in self.read_interface_infos:
            interfaces.append(rii.interface)
        return len(list(set(interfaces)))

    # Start raw logging for this interface
    def start_raw_logging(self):
        self.raw_logging_enabled = True
        self._change_raw_logging("start")

    # Stop raw logging for this interface
    def stop_raw_logging(self):
        self.raw_logging_enabled = False
        self._change_raw_logging("stop")

    # Supported Options
    # LISTEN_ADDRESS - Ip address of the interface to accept connections on - Default: 0.0.0.0
    # (see Interface#set_option)
    def set_option(self, option_name, option_values):
        super().set_option(option_name, option_values)
        match option_name.upper():
            case "LISTEN_ADDRESS":
                self.listen_address = option_values[0]

    def _shutdown_interfaces(self, interface_infos):
        with self.connection_mutex:
            for interface_info in interface_infos:
                interface_info.interface.disconnect()
                if interface_info.interface.stream_log_pair:
                    interface_info.interface.stream_log_pair.stop()
            interface_infos = []

    def _change_raw_logging(self, method):
        if self.stream_log_pair:
            for interface_info in self.write_interface_infos:
                if interface_info.interface.stream_log_pair:
                    getattr(interface_info.interface.stream_log_pair, method)
            for interface_info in self.read_interface_infos:
                if interface_info.interface.stream_log_pair:
                    getattr(interface_info.interface.stream_log_pair, method)

    def _start_listen_thread(self, port, listen_write=False, listen_read=False):
        listen_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
        listen_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            addr = (self.listen_address, port)
            listen_socket.bind(addr)
        except Exception:
            raise RuntimeError(
                f"Error binding to port {port}.\n"
                + "Either another application is using this port\n"
                + "or the operating system is being slow cleaning up.\n"
                + "Make sure all sockets/streams are closed in all applications,\n"
                + "wait 1 minute and try again."
            )

        listen_socket.setblocking(0)
        listen_socket.listen(5)
        self.listen_sockets.append(listen_socket)

        thread_reader, thread_writer = multiprocessing.Pipe()
        self.listen_pipes.append(thread_writer)
        thread = threading.Thread(
            target=self._listen_thread_body,
            args=[listen_socket, listen_write, listen_read, thread_reader],
            daemon=True,
        )
        thread.start()
        self.listen_threads.append(thread)

    def _listen_thread_body(self, listen_socket, listen_write, listen_read, thread_reader):
        while True:
            while True:
                try:
                    client_socket, address = listen_socket.accept()
                    break
                except (ConnectionAbortedError, BlockingIOError):
                    if self.cancel_threads:
                        break
                    # Wait for something to be readable
                    select.select([listen_socket, thread_reader], [], [])

            if self.cancel_threads:
                break
            host_ip, port = address
            try:
                hostname, _, _ = socket.gethostbyaddr(host_ip)
            except Exception:
                hostname = "UNKNOWN"

            # Configure TCP_NODELAY option
            client_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

            # Accept Connection
            write_socket = None
            read_socket = None
            if listen_write:
                write_socket = client_socket
            if listen_read:
                read_socket = client_socket
            stream = TcpipSocketStream(write_socket, read_socket, self.write_timeout, self.read_timeout)

            interface = StreamInterface()
            interface.target_names = self.target_names
            interface.cmd_target_names = self.cmd_target_names
            interface.tlm_target_names = self.tlm_target_names
            if self.stream_log_pair:
                interface.stream_log_pair = self.stream_log_pair.clone()
                if self.raw_logging_enabled:
                    interface.stream_log_pair.start()
            for protocol_class, protocol_args, read_write in self.protocol_info:
                interface.add_protocol(protocol_class, protocol_args, read_write)
            interface.stream = stream
            interface.connect()

            if listen_write:
                if self.write_connection_callback:
                    self.write_connection_callback.call(interface)
                with self.connection_mutex:
                    self.write_interface_infos.append(InterfaceInfo(interface, hostname, host_ip, port))
            if listen_read:
                if self.read_connection_callback:
                    self.read_connection_callback.call(interface)
                with self.connection_mutex:
                    self.read_interface_infos.append(InterfaceInfo(interface, hostname, host_ip, port))
                thread = threading.Thread(
                    target=self._start_read_thread,
                    args=[self.read_interface_infos[-1]],
                    daemon=True,
                )
                self.read_threads.append(thread)
                thread.start()
            Logger.info(f"{self.name}: Tcpip server accepted connection from {hostname}({host_ip}):{port}")

    def _start_read_thread(self, interface_info):
        try:
            try:
                self._read_thread_body(interface_info.interface)
            except Exception as error:
                Logger.error(f"{self.name}: Tcpip server read thread unexpectedly died")
                Logger.error(repr(error))
            Logger.info(
                f"{self.name}: Tcpip server lost read connection to {interface_info.hostname}({interface_info.host_ip}):{interface_info.port}"
            )
            self.read_threads.remove(threading.current_thread())

            index_to_delete = None
            with self.connection_mutex:
                index = 0
                for read_interface_info in self.read_interface_infos:
                    if interface_info.interface == read_interface_info.interface:
                        index_to_delete = index
                        read_interface_info.interface.disconnect()
                        if read_interface_info.interface.stream_log_pair:
                            read_interface_info.interface.stream_log_pair.stop()
                        break
                    index += 1
            if index_to_delete:
                del self.read_interface_infos[index_to_delete]
        except Exception as error:
            Logger.error(f"{self.name}: Tcpip server read thread unexpectedly died")
            Logger.error(repr(error))

    def _write_thread_body(self):
        try:
            while True:
                if self.cancel_threads:
                    break

                # Retrieve the next packet to be sent out to clients
                # Handles disconnected clients even when packets aren't flowing
                packet = None

                while True:
                    if self.cancel_threads:
                        break

                    try:
                        packet = self.write_queue.get_nowait()
                        break
                    except queue.Empty:
                        if self.cancel_threads:
                            break
                        self._check_for_dead_clients()

                packet = self._write_thread_hook(packet)
                if packet:
                    self._write_to_clients("write", packet)

        except Exception as error:
            self._shutdown_interfaces(self.write_interface_infos)
            Logger.error(f"{self.name}: Tcpip server write thread unexpectedly died")
            Logger.error(repr(error))

    def _write_raw_thread_body(self):
        try:
            while True:
                if self.cancel_threads:
                    break

                # Retrieve the next data to be sent out to clients
                data = None

                while True:
                    if self.cancel_threads:
                        break

                    try:
                        data = self.write_raw_queue.get_nowait()
                        break
                    except queue.Empty:
                        # Sleep until we receive data or for 100ms
                        with self.write_raw_condition_variable:
                            self.write_raw_condition_variable.wait(0.1)

                data = self._write_raw_thread_hook(data)
                if data:
                    self._write_to_clients("write_raw", data)

        except Exception as error:
            self._shutdown_interfaces(self.write_interface_infos)
            Logger.error(f"{self.name}: Tcpip server write raw thread unexpectedly died")
            Logger.error(repr(error))

    def _write_thread_hook(self, packet):
        return packet  # By default just return the packet

    def _write_raw_thread_hook(self, data):
        return data  # By default just return the data

    def _read_thread_body(self, interface):
        thread_bytes_read = 0
        while True:
            packet = interface.read()
            interface_bytes_read = interface.bytes_read
            if interface_bytes_read != thread_bytes_read:
                diff = interface_bytes_read - thread_bytes_read
                self.bytes_read += diff
                thread_bytes_read = interface_bytes_read
            if not packet or self.cancel_threads:
                return

            packet = self._read_thread_hook(packet)  # Do work on received packet
            self.read_raw_data_time = interface.read_raw_data_time
            self.read_raw_data = interface.read_raw_data
            self.read_queue.put(packet.clone())

    # @return [Packet] Return the packet
    def _read_thread_hook(self, packet):
        return packet

    def _check_for_dead_clients(self):
        indexes_to_delete = []
        index = 0

        with self.connection_mutex:
            try:
                for interface_info in self.write_interface_infos:
                    if self.write_port != self.read_port:
                        # Socket should return EWOULDBLOCK if it is still cleanly connected
                        interface_info.interface.stream.write_socket.recv(10, socket.MSG_DONTWAIT)
                    elif interface_info.interface.stream.write_socket.fileno() != -1:
                        # Let read thread detect disconnect
                        continue
                    # Client has disconnected (or is invalidly sending data on the socket)
                    Logger.info(
                        f"{self.name}: Tcpip server lost write connection to {interface_info.hostname}({interface_info.host_ip}):{interface_info.port}"
                    )
                    interface_info.interface.disconnect()
                    if interface_info.interface.stream_log_pair:
                        interface_info.interface.stream_log_pair.stop()
                    indexes_to_delete.insert(0, index)  # Put later indexes at front of array
            except socket.error as error:
                if error.errno == socket.EAGAIN or error.errno == socket.EWOULDBLOCK:
                    # Client is still cleanly connected as far as we can tell without writing to the socket
                    pass
                else:
                    # Client has disconnected
                    Logger.info(
                        f"{self.name}: Tcpip server lost write connection to {interface_info.hostname}({interface_info.host_ip}):{interface_info.port}"
                    )
                    interface_info.interface.disconnect()
                    if interface_info.interface.stream_log_pair:
                        interface_info.interface.stream_log_pair.stop()
                    indexes_to_delete.insert(0, index)  # Put later indexes at front of array
            finally:
                index += 1

            # Delete any dead sockets
            for index_to_delete in indexes_to_delete:
                del self.write_interface_infos[index_to_delete]

        # Sleep until we receive a packet or for 100ms
        with self.write_condition_variable:
            self.write_condition_variable.wait(0.1)

    def _write_to_clients(self, method, packet_or_data):
        with self.connection_mutex:
            # Send data to each client - On error drop the client
            indexes_to_delete = []
            index = 0
            for interface_info in self.write_interface_infos:
                need_disconnect = False
                try:
                    interface_bytes_written = interface_info.interface.bytes_written
                    getattr(interface_info.interface, method)(packet_or_data)
                    diff = interface_info.interface.bytes_written - interface_bytes_written
                    self.written_raw_data_time = interface_info.interface.written_raw_data_time
                    self.written_raw_data = interface_info.interface.written_raw_data
                    self.bytes_written += diff
                except IOError:
                    # Client has normally disconnected
                    need_disconnect = True
                except Exception as error:
                    Logger.error(f"{self.name}: Error sending to client: {error.__class__.__name__} {repr(error)}")
                    need_disconnect = True

                if need_disconnect:
                    Logger.info(
                        f"{self.name}: Tcpip server lost write connection to {interface_info.hostname}({interface_info.host_ip}):{interface_info.port}"
                    )
                    interface_info.interface.disconnect
                    if interface_info.interface.stream_log_pair:
                        interface_info.interface.stream_log_pair.stop
                    indexes_to_delete.insert(0, index)  # Put later indexes at front of array
                index += 1

            # Delete any dead sockets
            for index_to_delete in indexes_to_delete:
                del self.write_interface_infos[index_to_delete]



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import socket
from openc3.io.udp_sockets import UdpReadSocket, UdpWriteSocket, UdpReadWriteSocket
from openc3.interfaces.interface import Interface
from openc3.config.config_parser import ConfigParser
from openc3.utilities.logger import Logger
from openc3.utilities.sleeper import Sleeper
from openc3.top_level import close_socket


# Base class for interfaces that send and receive messages over UDP
class UdpInterface(Interface):
    # @param hostname [String] Machine to connect to
    # @param write_dest_port [Integer] Port to write commands to
    # @param read_port [Integer] Port to read telemetry from
    # @param write_src_port [Integer] Port to allow replies if needed
    # @param interface_address [String] If the destination machine represented
    #   by hostname supports multicast, then interface_address is used to
    #   configure the outgoing multicast address.
    # @param ttl [Integer] Time To Live value. The number of intermediate
    #   routers allowed before dropping the packet.
    # @param write_timeout [Float] Seconds to wait before aborting writes
    # @param read_timeout [Float|nil] Seconds to wait before aborting reads.
    #   Pass nil to block until the read is complete.
    # @param bind_address [String] Address to bind UDP ports to
    def __init__(
        self,
        hostname,
        write_dest_port,
        read_port,
        write_src_port=None,
        interface_address=None,
        ttl=128,  # default for Windows
        write_timeout=10.0,
        read_timeout=None,
        bind_address="0.0.0.0",
    ):
        super().__init__()
        self.hostname = ConfigParser.handle_none(hostname)
        if self.hostname is not None:
            self.hostname = str(hostname)
            if self.hostname.upper() == "LOCALHOST":
                self.hostname = "127.0.0.1"
        self.write_dest_port = ConfigParser.handle_none(write_dest_port)
        if self.write_dest_port is not None:
            self.write_dest_port = int(write_dest_port)
        self.read_port = ConfigParser.handle_none(read_port)
        if self.read_port is not None:
            self.read_port = int(read_port)
        self.write_src_port = ConfigParser.handle_none(write_src_port)
        if self.write_src_port is not None:
            self.write_src_port = int(write_src_port)
        self.interface_address = ConfigParser.handle_none(interface_address)
        if self.interface_address and self.interface_address.upper() == "LOCALHOST":
            self.interface_address = "127.0.0.1"
        self.ttl = int(ttl)
        if self.ttl < 1:
            self.ttl = 1
        self.write_timeout = ConfigParser.handle_none(write_timeout)
        if self.write_timeout is not None:
            self.write_timeout = float(write_timeout)
        else:
            Logger.warn("Warning: To avoid interface lock, write_timeout can not be None. Setting to 10 seconds.")
            self.write_timeout = 10.0
        self.read_timeout = ConfigParser.handle_none(read_timeout)
        if self.read_timeout is not None:
            self.read_timeout = float(read_timeout)
        self.bind_address = ConfigParser.handle_none(bind_address)
        if self.bind_address and self.bind_address.upper() == "LOCALHOST":
            self.bind_address = "127.0.0.1"
        self.write_socket = None
        self.read_socket = None
        if self.read_port is None:
            self.read_allowed = False
        if self.write_dest_port is None:
            self.write_allowed = False
        if self.write_dest_port is None:
            self.write_raw_allowed = False

    def connection_string(self):
        result = ""
        if self.write_dest_port:
            result += f" {self.hostname}:{self.write_dest_port} (write dest port)"
        if self.write_src_port:
            result += f" {self.write_src_port} (write src port)"
        if self.read_port:
            result += f" {self.hostname}:{self.read_port} (read)"
        if self.interface_address:
            result += f" {self.interface_address} (interface addr)"
        if self.bind_address != "0.0.0.0":
            result += f" {self.bind_address} (bind addr)"
        return result.strip()

    # Creates a new {UdpWriteSocket} if the the write_dest_port was given in
    # the constructor and a new {UdpReadSocket} if the read_port was given in
    # the constructor.
    def connect(self):
        if self.read_port and self.write_dest_port and self.write_src_port and (self.read_port == self.write_src_port):
            self.read_socket = UdpReadWriteSocket(
                self.read_port,
                self.bind_address,
                self.write_dest_port,
                self.hostname,
                self.interface_address,
                self.ttl,
            )
            self.write_socket = self.read_socket
        else:
            if self.read_port:
                self.read_socket = UdpReadSocket(
                    self.read_port,
                    self.hostname,
                    self.interface_address,
                    self.bind_address,
                )
            if self.write_dest_port:
                self.write_socket = UdpWriteSocket(
                    self.hostname,
                    self.write_dest_port,
                    self.write_src_port,
                    self.interface_address,
                    self.ttl,
                    self.bind_address,
                )
        self.thread_sleeper = None
        super().connect()

    # @return [Boolean] Whether the active ports (read and/or write) have
    #   created sockets. Since UDP is connectionless, creation of the sockets
    #   is used to determine connection.
    def connected(self):
        if self.write_dest_port is not None and self.read_port is not None:
            if self.write_socket and self.read_socket:
                return True
            else:
                return False
        elif self.write_dest_port is not None:
            if self.write_socket:
                return True
            else:
                return False
        else:
            if self.read_socket:
                return True
            else:
                return False

    # Close the active ports (read and/or write) and set the sockets to nil.
    def disconnect(self):
        if self.write_socket != self.read_socket:
            close_socket(self.write_socket)
        close_socket(self.read_socket)
        self.write_socket = None
        self.read_socket = None
        if self.thread_sleeper:
            self.thread_sleeper.cancel()
        self.thread_sleeper = None
        super().disconnect()

    def read(self):
        if self.read_port:
            return super().read()

        # Write only interface so stop the thread which calls read
        self.thread_sleeper = Sleeper()
        while self.connected():
            self.thread_sleeper.sleep(1_000_000_000)
        return None

    # Reads from the socket if the read_port is defined
    def read_interface(self):
        try:
            data = self.read_socket.read(self.read_timeout)
            if len(data) <= 0:
                Logger.info(f"{self.name}: Udp read returned 0 bytes (stream closed)")
            extra = None
            self.read_interface_base(data, extra)
            return (data, extra)
        # TODO: select.select can throw TypeErorr: fileno() returned a non-integer
        # Does it also throw socket.error?
        except (TypeError, socket.error):
            return None, None

    # Writes to the socket
    # @param data [String] Raw packet data
    def write_interface(self, data, extra=None):
        self.write_interface_base(data, extra)
        self.write_socket.write(data, self.write_timeout)
        return data, extra



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import copy
import threading
from contextlib import contextmanager
from openc3.accessors.binary_accessor import BinaryAccessor
from openc3.packets.structure_item import StructureItem
from openc3.utilities.string import formatted


class Structure:
    """Maintains knowledge of a raw binary structure. Uses structure_item to
    create individual structure items which are read and written by
    binary_accessor."""

    # String providing a single 0 byte
    ZERO_STRING = b"\00"

    def __init__(
        self,
        default_endianness=BinaryAccessor.HOST_ENDIANNESS,
        buffer=None,
        item_class=StructureItem,
    ):
        if (default_endianness == "BIG_ENDIAN") or (default_endianness == "LITTLE_ENDIAN"):
            self.default_endianness = default_endianness
            if buffer is not None and not isinstance(buffer, (bytes, bytearray)):
                raise TypeError(f"wrong argument type {buffer.__class__.__name__} (expected bytes)")
            if buffer is None:
                self._buffer = None
            else:
                self._buffer = bytearray(buffer)  # TODO: Do we need to force encoding?
            self.item_class = item_class
            self.items = {}
            self.sorted_items = []
            self.defined_length = 0
            self.defined_length_bits = 0
            self.pos_bit_size = 0
            self.neg_bit_size = 0
            self.fixed_size = True
            self.short_buffer_allowed = False
            self.mutex = None
            self.accessor = BinaryAccessor(self)
        else:
            raise ValueError(f"Unknown endianness '{default_endianness}', must be 'BIG_ENDIAN' or 'LITTLE_ENDIAN'")

    # Read an item in the structure
    #
    # self.param item [StructureItem] Instance of StructureItem or one of its subclasses
    # self.param value_type [Symbol] Not used. Subclasses should overload this
    #   parameter to check whether to perform conversions on the item.
    # self.param buffer [String] The binary buffer to read the item from
    # self.return Value based on the item definition. This could be a string, integer,
    #   float, or array of values.
    def read_item(self, item, value_type="RAW", buffer=None):
        if not buffer:
            buffer = self._buffer
        if not buffer:
            buffer = self.allocate_buffer_if_needed()
        return self.accessor.read_item(item, buffer)

    # Get the length of the buffer used by the structure
    #
    # self.return [Integer] Size of the buffer in bytes
    def length(self):
        self.allocate_buffer_if_needed()
        return len(self._buffer)

    # Resize the buffer at least the defined length of the structure
    def resize_buffer(self):
        if self._buffer:
            # Extend data size
            if len(self._buffer) < self.defined_length:
                self._buffer += Structure.ZERO_STRING * (self.defined_length - len(self._buffer))
        else:
            self.allocate_buffer_if_needed()

    @property
    def accessor(self):
        return self.__accessor

    # Configure the accessor for this packet
    #
    # self.param accessor [Accessor] The class to use as an accessor
    @accessor.setter
    def accessor(self, accessor):
        self.__accessor = accessor
        # isinstance can fail if the class is reloaded because the class becomes a new class
        # so directly check the class name which is basically equivalent
        if self.__accessor.enforce_short_buffer_allowed():
            self.short_buffer_allowed = True

    # Read a list of items in the structure
    #
    # self.param items [StructureItem] Array of StructureItem or one of its subclasses
    # self.param value_type [Symbol] Not used. Subclasses should overload this
    #   parameter to check whether to perform conversions on the item.
    # self.param buffer [String] The binary buffer to read the item from
    # self.return Hash of read names and values
    def read_items(self, items, value_type="RAW", buffer=None):
        if not buffer:
            buffer = self._buffer
        if not buffer:
            buffer = self.allocate_buffer_if_needed()
        return self.accessor.read_items(items, buffer)

    # Allocate a buffer if not available
    def allocate_buffer_if_needed(self):
        if not self._buffer:
            self._buffer = bytearray(Structure.ZERO_STRING * self.defined_length)
        return self._buffer

    # Indicates if any items have been defined for this structure
    # self.return [TrueClass or FalseClass]
    def defined(self):
        return len(self.sorted_items) > 0

    # Rename an existing item
    #
    # self.param item_name [String] Name of the currently defined item
    # self.param new_item_name [String] New name for the item
    def rename_item(self, item_name, new_item_name):
        item = self.get_item(item_name)
        item.name = new_item_name
        self.items.pop(item_name)
        self.items[new_item_name] = item
        # Since self.sorted_items contains the actual item reference it is
        # updated when we set the item.name
        return item

    # Define an item in the structure. This creates a new instance of the
    # item_class as given in the constructor and adds it to the items hash. It
    # also resizes the buffer to accommodate the new item.
    #
    # self.param name [String] Name of the item. Used by the items hash to retrieve
    #   the item.
    # self.param bit_offset [Integer] Bit offset of the item in the raw buffer
    # self.param bit_size [Integer] Bit size of the item in the raw buffer
    # self.param data_type [Symbol] Type of data contained by the item. This is
    #   dependent on the item_class but by default see StructureItem.
    # self.param array_size [Integer] Set to a non None value if the item is to
    #   represented as an array.
    # self.param endianness [Symbol] Endianness of this item. By default the
    #   endianness as set in the constructor is used.
    # self.param overflow [Symbol] How to handle value overflows. This is
    #   dependent on the item_class but by default see StructureItem.
    # self.return [StrutureItem] The structure item defined
    def define_item(
        self,
        name,
        bit_offset,
        bit_size,
        data_type,
        array_size=None,
        endianness=None,
        overflow="ERROR",
    ):
        if not endianness:
            endianness = self.default_endianness
        # Create the item
        item = self.item_class(name, bit_offset, bit_size, data_type, endianness, array_size, overflow)
        return self.define(item)

    # Adds the given item to the items hash. It also resizes the buffer to
    # accommodate the new item.
    #
    # self.param item [StructureItem] The structure item to add
    # self.return [StrutureItem] The structure item defined
    def define(self, item):
        # Handle Overwriting Existing Item
        if self.items.get(item.name):
            item_index = None
            for index, sorted_item in enumerate(self.sorted_items):
                if sorted_item.name == item.name:
                    item_index = index
                    break
            if item_index < len(self.sorted_items):
                self.sorted_items.pop(item_index)

        # Add to Sorted Items
        if len(self.sorted_items) != 0:
            last_item = self.sorted_items[-1]
            self.sorted_items.append(item)
            # If the current item or last item have a negative offset then we have
            # to re-sort. We also re-sort if the current item is less than the last
            # item because we are inserting.
            if last_item.bit_offset <= 0 or item.bit_offset <= 0 or item.bit_offset < last_item.bit_offset:
                self.sorted_items.sort()
        else:
            self.sorted_items.append(item)

        # Add to the overall hash of defined items
        self.items[item.name] = item
        # Update fixed size knowledge
        if (item.data_type != "DERIVED" and item.bit_size <= 0) or (item.array_size and item.array_size <= 0):
            self.fixed_size = False

        # Recalculate the overall defined length of the structure
        update_needed = False
        if item.bit_offset >= 0:
            if item.bit_size > 0:
                if item.array_size is not None:
                    if item.array_size >= 0:
                        item_defined_length_bits = item.bit_offset + item.array_size
                    else:
                        item_defined_length_bits = item.bit_offset
                else:
                    item_defined_length_bits = item.bit_offset + item.bit_size

                if item_defined_length_bits > self.pos_bit_size:
                    self.pos_bit_size = item_defined_length_bits
                    update_needed = True

            elif item.bit_offset > self.pos_bit_size:
                self.pos_bit_size = item.bit_offset
                update_needed = True

        else:
            if abs(item.bit_offset) > self.neg_bit_size:
                self.neg_bit_size = abs(item.bit_offset)
                update_needed = True

        if update_needed:
            self.defined_length_bits = self.pos_bit_size + self.neg_bit_size
            self.defined_length = int(self.defined_length_bits / 8)
            if self.defined_length_bits % 8 != 0:
                self.defined_length += 1

        # Resize the buffer if necessary
        if self.buffer:
            self.resize_buffer()
        return item

    # Define an item at the end of the structure. This creates a new instance of the
    # item_class as given in the constructor and adds it to the items hash. It
    # also resizes the buffer to accommodate the new item.
    #
    # self.param name (see #define_item)
    # self.param bit_size (see #define_item)
    # self.param data_type (see #define_item)
    # self.param array_size (see #define_item)
    # self.param endianness (see #define_item)
    # self.param overflow (see #define_item)
    # self.return (see #define_item)
    def append_item(
        self,
        name,
        bit_size,
        data_type,
        array_size=None,
        endianness=None,
        overflow="ERROR",
    ):
        if not endianness:
            endianness = self.default_endianness
        if data_type == "DERIVED":
            return self.define_item(name, 0, bit_size, data_type, array_size, endianness, overflow)
        else:
            return self.define_item(
                name,
                self.defined_length_bits,
                bit_size,
                data_type,
                array_size,
                endianness,
                overflow,
            )

    # Adds an item at the  of the structure. It adds the item to the items
    # hash and resizes the buffer to accommodate the new item.
    #
    # self.param item (see #define)
    # self.return (see #define)
    def append(self, item):
        if item.data_type == "DERIVED":
            item.bit_offset = 0
        else:
            # We're appending a new item so set the bit_offset
            item.bit_offset = self.defined_length_bits
            # Also set original_bit_offset because it's currently 0
            # due to PacketItemParser::create_packet_item
            # get_bit_offset() returning 0 if append
            item.original_bit_offset = self.defined_length_bits

        return self.define(item)

    # self.param name [String] Name of the item to look up in the items Hash
    # self.return [StructureItem] StructureItem or one of its subclasses
    def get_item(self, name):
        item = self.items.get(name.upper())
        if not item:
            raise ValueError(f"Unknown item: {name}")
        return item

    # self.param item [#name] Instance of StructureItem or one of its subclasses.
    #   The name method will be used to look up the item and set it to the new instance.
    def set_item(self, item):
        if self.items.get(item.name):
            self.items[item.name] = item

            # Need to allocate space for the variable length item if its minimum size is greater than zero
            if item.variable_bit_size:
                minimum_data_bits = 0
                if (item.data_type == "INT" or item.data_type == "UINT") and not item.original_array_size:
                    # Minimum QUIC encoded integer, see https://datatracker.ietf.org/doc/html/rfc9000#name-variable-length-integer-enc
                    minimum_data_bits = 6
                # STRING, BLOCK, or array item
                elif item.variable_bit_size["length_value_bit_offset"] > 0:
                    minimum_data_bits = (
                        item.variable_bit_size["length_value_bit_offset"]
                        * item.variable_bit_size["length_bits_per_count"]
                    )
                if minimum_data_bits > 0 and item.bit_offset >= 0 and self.defined_length_bits == item.bit_offset:
                    self.defined_length_bits += minimum_data_bits
        else:
            raise ValueError(f"Unknown item: {item.name} - Ensure item name is uppercase")

    # self.param name [String] Name of the item to delete in the items Hash
    def delete_item(self, name):
        item = self.items[name.upper()]
        if not item:
            raise RuntimeError(f"Unknown item: {name}")

        # Find the item to delete in the sorted_items array
        item_index = None
        for index, sorted_item in enumerate(self.sorted_items):
            if sorted_item.name == item.name:
                item_index = index
                break

        self.sorted_items.pop(item_index)
        self.items.pop(name.upper())

    # Write a value to the buffer based on the item definition
    #
    # self.param item [StructureItem] Instance of StructureItem or one of its subclasses
    # self.param value [Object] Value based on the item definition. This could be
    #   a string, integer, float, or array of values.
    # self.param value_type [Symbol] Not used. Subclasses should overload this
    #   parameter to check whether to perform conversions on the item.
    # self.param buffer [String] The binary buffer to write the value to
    def write_item(self, item, value, value_type="RAW", buffer=None):
        if not buffer:
            buffer = self._buffer
        if not buffer:
            buffer = self.allocate_buffer_if_needed()
        self.accessor.write_item(item, value, buffer)

    # Write values to the buffer based on the item definitions
    #
    # self.param items [StructureItem] Array of StructureItem or one of its subclasses
    # self.param value [Object] Array of values based on the item definitions.
    # self.param value_type [Symbol] Not used. Subclasses should overload this
    #   parameter to check whether to perform conversions on the item.
    # self.param buffer [String] The binary buffer to write the values to
    def write_items(self, items, values, value_type="RAW", buffer=None):
        if not buffer:
            buffer = self._buffer
        if not buffer:
            buffer = self.allocate_buffer_if_needed()
        self.accessor.write_items(items, values, buffer)

    # Read an item in the structure by name
    #
    # self.param name [String] Name of an item to read
    # self.param value_type [Symbol] Not used. Subclasses should overload this
    #   parameter to check whether to perform conversions on the item.
    # self.param buffer [String] The binary buffer to read the item from
    # self.return Value based on the item definition. This could be an integer,
    #   float, or array of values.
    def read(self, name, value_type="RAW", buffer=None):
        if not buffer:
            buffer = self._buffer
        return self.read_item(self.get_item(name), value_type, buffer)

    # Write an item in the structure by name
    #
    # self.param name [Object] Name of the item to write
    # self.param value [Object] Value based on the item definition. This could be
    #   a string, integer, float, or array of values.
    # self.param value_type [Symbol] Not used. Subclasses should overload this
    #   parameter to check whether to perform conversions on the item.
    # self.param buffer [String] The binary buffer to write the value to
    def write(self, name, value, value_type="RAW", buffer=None):
        if not buffer:
            buffer = self._buffer
        self.write_item(self.get_item(name), value, value_type, buffer)

    # Read all items in the structure into an array of arrays
    #   [[item name, item value], ...]
    #
    # self.param value_type [Symbol] Not used. Subclasses should overload this
    #   parameter to check whether to perform conversions on the item.
    # self.param buffer [String] The binary buffer to write the value to
    # self.param top [Boolean] Indicates if this is a top level call for the mutex
    # self.return [Array<Array>] Array of two element arrays containing the item
    #   name as element 0 and item value as element 1.
    def read_all(self, value_type="RAW", buffer=None, top=True):
        if not buffer:
            buffer = self._buffer
        item_array = []
        with self.synchronize_allow_reads(top):
            for item in self.sorted_items:
                item_array.append([item.name, self.read_item(item, value_type, buffer)])
        return item_array

    # Create a string that shows the name and value of each item in the structure
    #
    # self.param value_type [Symbol] Not used. Subclasses should overload this
    #   parameter to check whether to perform conversions on the item.
    # self.param indent [Integer] Amount to indent before printing the item name
    # self.param buffer [String] The binary buffer to write the value to
    # self.param ignored [Array<String>] List of items to ignore when building the string
    # self.return [String] String formatted with all the item names and values
    def formatted(self, value_type="RAW", indent=0, buffer=None, ignored=None):
        if not buffer:
            buffer = self._buffer
        indent_string = " " * indent
        string = ""
        with self.synchronize_allow_reads(True):
            for item in self.sorted_items:
                if ignored and item.name in ignored:
                    continue

                if (item.data_type != "BLOCK") or (
                    item.data_type == "BLOCK" and value_type != "RAW" and hasattr(item, "read_conversion")
                ):
                    string += f"{indent_string}{item.name}: {self.read_item(item, value_type, buffer)}\n"
                else:
                    value = self.read_item(item, value_type, buffer)
                    if isinstance(value, (str, bytes, bytearray)):
                        string += f"{indent_string}{item.name}:\n"
                        string += formatted(value, 1, 16, " ", indent + 2)
                    else:
                        string += f"{indent_string}{item.name}: {value}\n"

        return string

    # Get the buffer used by the structure. The current buffer is copied and
    # thus modifications to the returned buffer will have no effect on the
    # structure items.
    #
    # self.param copy [TrueClass/FalseClass] Whether to copy the buffer
    # self.return [String] Data buffer backing the structure
    @property
    def buffer(self):
        return self.allocate_buffer_if_needed()[:]

    def buffer_no_copy(self):
        return self.allocate_buffer_if_needed()

    # Set the buffer to be used by the structure. The buffer is copied and thus
    # further modifications to the buffer have no effect on the structure
    # items.
    #
    # self.param buffer [String] Buffer of data to back the structure items
    @buffer.setter
    def buffer(self, buffer):
        with self.synchronize():
            self.internal_buffer_equals(buffer)

    # Make a light weight clone of this structure. This only creates a new buffer
    # of data. The defined structure items are the same.
    #
    # self.return [Structure] A copy of the current structure with a new underlying
    #   buffer of data
    def clone(self):
        struct = copy.copy(self)
        struct._buffer = self.buffer  # Makes a copy
        struct.accessor = copy.copy(self.accessor)
        struct.accessor.packet = struct
        return struct

    # Clone that also deep copies items
    # @return [Structure] A deep copy of the structure
    def deep_copy(self):
        cloned = self.clone()
        cloned_items = []
        for item in cloned.sorted_items:
            cloned_items.append(item.clone())

        cloned.sorted_items = cloned_items
        cloned.items = {}
        for item in cloned_items:
            cloned.items[item.name] = item
        return cloned

    CLASS_MUTEX = threading.Lock()

    def setup_mutex(self):
        if self.mutex:
            return
        with Structure.CLASS_MUTEX:
            self.mutex_allow_reads = False
            self.mutex = threading.Lock()

    # Take the structure mutex to ensure the buffer does not change while you perform activities
    def synchronize(self):
        self.setup_mutex()
        return self.mutex

    # Take the structure mutex to ensure the buffer does not change while you perform activities
    # This versions allows reads to happen if a top level function has already taken the mutex
    # self.param top [Boolean] If True this will take the mutex and set an allow reads flag to allow
    #      lower level calls to go forward without getting the mutex
    @contextmanager
    def synchronize_allow_reads(self, top=False):
        self.setup_mutex()
        if top:
            with self.mutex:
                self.mutex_allow_reads = threading.get_ident()
                try:
                    yield
                finally:
                    self.mutex_allow_reads = False
        else:
            got_mutex = self.mutex.acquire(False)
            if got_mutex:
                try:
                    yield
                finally:
                    self.mutex.release()
            else:  # if self.mutex_allow_reads == threading.get_ident()
                yield

    def calculate_total_bit_size(self, item):
        if item.variable_bit_size:
            # Bit size is determined by length field
            length_value = self.read(item.variable_bit_size["length_item_name"], "CONVERTED")
            if item.data_type == "INT" or item.data_type == "UINT" and not item.original_array_size:
                match length_value:
                    case 0:
                        return 6
                    case 1:
                        return 14
                    case 2:
                        return 30
                    case _:
                        return 62
            else:
                return (length_value * item.variable_bit_size["length_bits_per_count"]) + item.variable_bit_size[
                    "length_value_bit_offset"
                ]
        elif item.original_bit_size <= 0:
            # Bit size is full packet length - bits before item + negative bits saved at end
            return (len(self._buffer) * 8) - item.bit_offset + item.original_bit_size
        elif item.original_array_size and item.original_array_size <= 0:
            # Bit size is full packet length - bits before item + negative bits saved at end
            return (len(self._buffer) * 8) - item.bit_offset + item.original_array_size
        else:
            raise RuntimeError("Unexpected use of calculate_total_bit_size for non-variable-sized item")

    def recalculate_bit_offsets(self):
        adjustment = 0
        for item in self.sorted_items:
            # Anything with a negative bit offset should be left alone
            if item.original_bit_offset >= 0:
                item.bit_offset = item.original_bit_offset + adjustment
                if item.data_type != "DERIVED" and (
                    item.variable_bit_size
                    or item.original_bit_size <= 0
                    or (item.original_array_size and item.original_array_size <= 0)
                ):
                    new_bit_size = self.calculate_total_bit_size(item)
                    if item.original_bit_size != new_bit_size:
                        adjustment += new_bit_size - item.original_bit_size

    def internal_buffer_equals(self, buffer):
        if not isinstance(buffer, (bytes, bytearray)):
            raise TypeError(f"Buffer class is {buffer.__class__.__name__} but must be bytearray")

        self._buffer = bytearray(buffer[:])
        if not self.fixed_size:
            self.recalculate_bit_offsets()

        if self.accessor.enforce_length():
            if len(self._buffer) != self.defined_length:
                if len(self._buffer) < self.defined_length:
                    self.resize_buffer()
                    if not self.short_buffer_allowed:
                        raise ValueError("Buffer length less than defined length")
                elif self.fixed_size and self.defined_length != 0:
                    raise ValueError("Buffer length greater than defined length")



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import json
from jsonpath_ng import parse
from .accessor import Accessor


class JsonAccessor(Accessor):
    @classmethod
    def class_read_item(cls, item, buffer):
        if item.data_type == "DERIVED":
            return None
        if isinstance(buffer, bytearray):
            buffer = json.loads(buffer.decode())
        result = parse(item.key).find(buffer)
        if len(result) == 0:
            return None
        return cls.convert_to_type(result[0].value, item)

    @classmethod
    def class_write_item(cls, item, value, buffer):
        if item.data_type == "DERIVED":
            return None
        if isinstance(buffer, bytearray):
            decoded = json.loads(buffer.decode())
        else:
            decoded = buffer

        value = cls.convert_to_type(value, item)
        jsonpath = parse(item.key)
        # update_or_create doesn't always work if the item is there
        # so first find the item before updating
        matches = jsonpath.find(decoded)
        if len(matches) == 0:
            jsonpath.update_or_create(decoded, value)
        else:
            jsonpath.update(decoded, value)

        if isinstance(buffer, bytearray):
            # buffer[0:] syntax so we copy into the buffer
            buffer[0:] = bytearray(json.dumps(decoded), encoding="utf-8")

    @classmethod
    def class_read_items(cls, items, buffer):
        if isinstance(buffer, bytearray):
            buffer = json.loads(buffer.decode())
        return super().class_read_items(items, buffer)

    @classmethod
    def class_write_items(cls, items, values, buffer):
        if isinstance(buffer, bytearray):
            decoded = json.loads(buffer.decode())
        else:
            decoded = buffer

        super().class_write_items(items, values, decoded)
        if isinstance(buffer, bytearray):
            # buffer[0:] syntax so we copy into the buffer
            buffer[0:] = bytearray(json.dumps(decoded), encoding="utf-8")

    def enforce_encoding(self):
        return None

    def enforce_length(self):
        return False

    def enforce_short_buffer_allowed(self):
        return True

    def enforce_derived_write_conversion(self, item):
        return True



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import time
import json
import typing

from openc3.utilities.extract import *
import openc3.script
from openc3.environment import OPENC3_SCOPE


def metadata_all(limit: int = 100, scope: str = OPENC3_SCOPE):
    """Gets all the metadata

    Args:
        limit (int) Optional, defaults to 100
        scope (str) Optional, defaults to env.OPENC3_SCOPE

    Return:
        The json result of the method call
    """
    response = openc3.script.API_SERVER.request("get", "/openc3-api/metadata", query={"limit": limit}, scope=scope)
    # Non-existent just returns None
    if not response or response.status_code != 200:
        return None
    return json.loads(response.text)


def metadata_get(start: typing.Optional[int] = None, scope: str = OPENC3_SCOPE):
    """Gets metadata, default is latest if start is None

    Args:
        start (int) Metadata time value as integer seconds from epoch, Optional, defaults to "latest"
        scope (str) Optional, defaults to env.OPENC3_SCOPE

    Return:
        The json result of the method call
    """
    start_str = f"{start}" if start else "latest"
    response = openc3.script.API_SERVER.request("get", f"/openc3-api/metadata/{start_str}", scope=scope)

    # Non-existent just returns None
    if not response or response.status_code != 200:
        return None
    return json.loads(response.text)


def metadata_set(
    metadata: dict, start: typing.Optional[int] = None, color: typing.Optional[str] = None, scope: str = OPENC3_SCOPE
):
    """Create a new metadata entry at the given start time or now if no start given

    Args:
        metadata (dict) A dict of metadata
        start (int) Metadata time value as integer seconds from epoch
        color (str) Events color to show on Calendar tool, if None will be blue
        scope (str) Optional, defaults to env.OPENC3_SCOPE

    Return:
        The json result of the method call
    """
    if not isinstance(metadata, dict):
        raise TypeError(f"metadata must be a dict: {metadata} is a {metadata.__class__.__name__}")

    data = {"color": color if color else "#003784", "metadata": metadata}
    if start:
        data["start"] = time.asctime(time.gmtime(start))
    response = openc3.script.API_SERVER.request("post", "/openc3-api/metadata", data=data, json=True, scope=scope)
    if not response:
        raise RuntimeError(f"Failed to set metadata due to {response.status_code}")
    elif response.status_code == 409:
        raise RuntimeError("Metadata overlaps existing metadata. Did you metadata_set within 1s of another?")
    elif response.status_code != 201:
        raise RuntimeError(f"Failed to set metadata due to {response.status_code}")
    return json.loads(response.text)


def metadata_update(
    metadata: dict, start: typing.Optional[int] = None, color: typing.Optional[str] = None, scope: str = OPENC3_SCOPE
):
    """Updates existing metadata. If no start is given, updates latest metadata.

    Args:
        metadata (Dict[str, Any]) A dict of metadata
        start (int) Metadata time value as integer seconds from epoch
        color (str) Events color to show on Calendar tool, if None will be blue
        scope (str) Optional, defaults to env.OPENC3_SCOPE

    Return:
        The json result of the method call
    """
    if not isinstance(metadata, dict):
        raise TypeError(f"metadata must be a Hash: {metadata} is a {metadata.__class__.__name__}")

    if start is None:  # No start so grab latest
        existing = metadata_get()
        start = existing["start"]
        if not color:
            color = existing["color"]
        metadata = existing["metadata"] | metadata
    else:
        if not color:
            color = "#003784"

    data = {"color": color, "metadata": metadata, "start": time.asctime(time.gmtime(start))}
    response = openc3.script.API_SERVER.request(
        "put", f"/openc3-api/metadata/{start}", data=data, json=True, scope=scope
    )
    if not response or response.status_code != 200:
        raise RuntimeError("Failed to update metadata")

    return json.loads(response.text)


# Requests the metadata from the user for a target
def metadata_input(*args, **kwargs):
    raise RuntimeError("can only be used in Script Runner")



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import copy
from datetime import datetime, timezone
from .log_writer import LogWriter
from openc3.environment import OPENC3_SCOPE
from openc3.utilities.time import to_nsec_from_epoch
from openc3.utilities.logger import Logger


# Creates a log file of stream data for either reads or writes. Can automatically
# cycle the log based on when the log file reaches a predefined size or based on time.
class StreamLog(LogWriter):
    # # self.return [String] Original name passed to stream log
    # attr_reader :orig_name

    # The allowable log types
    LOG_TYPES = ["READ", "WRITE"]

    # self.param log_name [String] The name of the stream log. Typically matches the
    #    name of the corresponding interface
    # self.param log_type [Symbol] The type of log to create. Must be 'READ'
    #   or 'WRITE'.
    # self.param cycle_time [Integer] The amount of time in seconds before creating
    #   a new log file. This can be combined with cycle_size.
    # self.param cycle_size [Integer] The size in bytes before creating a new log
    #   file. This can be combined with cycle_time.
    # self.param cycle_hour [Integer] The time at which to cycle the log. Combined with
    #   cycle_minute to cycle the log daily at the specified time. If None, the log
    #   will be cycled hourly at the specified cycle_minute.
    # self.param cycle_minute [Integer] The time at which to cycle the log. See cycle_hour
    #   for more information.
    def __init__(
        self,
        log_name,
        log_type,
        cycle_time=600,  # 10 minutes, matches time in target_model
        cycle_size=50_000_000,  # 50MB, matches size in target_model
        cycle_hour=None,
        cycle_minute=None,
    ):
        if log_type not in StreamLog.LOG_TYPES:
            raise RuntimeError("log_type must be 'READ' or 'WRITE'")

        super().__init__(
            f"{OPENC3_SCOPE}/stream_logs/",
            True,  # Start with logging enabled
            cycle_time,
            cycle_size,
            cycle_hour,
            cycle_minute,
        )

        self.log_type = log_type
        self.name = log_name

    @property
    def name(self):
        return self.log_name

    @name.setter
    def name(self, name):
        self.orig_name = name
        self.log_name = name.lower() + "_stream_" + self.log_type.lower()

    # Create a clone of this object with a new name
    def clone(self):
        stream_log = copy.copy(self)
        stream_log.name = stream_log.orig_name
        return stream_log

    # Write to the log file.
    #
    # If no log file currently exists in the filesystem, a new file will be
    # created.
    #
    # self.param data [String] String of data
    def write(self, data):
        if not self.logging_enabled:
            return
        if data is None or len(data) <= 0:
            return

        try:
            with self.mutex:
                time_nsec_since_epoch = to_nsec_from_epoch(datetime.now(timezone.utc))
                self.prepare_write(time_nsec_since_epoch, len(data))
                if self.file:
                    self.write_entry(time_nsec_since_epoch, data)
        except RuntimeError as error:
            Logger.error(f"Error writing {self.filename} : {repr(error)}")
            # OpenC3.handle_critical_exception(err)

    def write_entry(self, time_nsec_since_epoch, data):
        self.file.write(data)
        self.file_size += len(data)
        if not self.first_time:
            self.first_time = time_nsec_since_epoch
        self.last_time = time_nsec_since_epoch

    def bucket_filename(self):
        return f"{self.first_timestamp()}__{self.log_name}" + self.extension()

    def extension(self):
        return ".bin"



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import os
import sys
import threading
import importlib
import time
import socket
import traceback
from openc3.utilities.logger import Logger

openc3_chdir_mutex = threading.RLock()


class HazardousError(Exception):
    def __init__(self):
        self.target_name = ""
        self.cmd_name = ""
        self.cmd_params = ""
        self.hazardous_description = ""
        self.formatted = ""
        super().__init__()

    def __str__(self):
        string = f"{self.target_name} {self.cmd_name} with {self.cmd_params} is Hazardous"
        if self.hazardous_description:
            string += f"due to '{self.hazardous_description}'"
        # Pass along the original formatted command so it can be resent
        string += f".\n{self.formatted}"
        return string


class CriticalCmdError(Exception):
    def __init__(self):
        self.uuid = ""
        self.username = ""
        self.target_name = ""
        self.cmd_name = ""
        self.cmd_params = {}
        self.cmd_string = ""
        super().__init__()


class DisabledError(Exception):
    def __init__(self):
        self.target_name = ""
        self.cmd_name = ""
        super().__init__()

    def __str__(self):
        return f"{self.target_name} {self.cmd_name} is Disabled"


# Adds a path to the global Python search path
#
# @param path [String] Directory path
def add_to_search_path(path, front=True):
    path = os.path.abspath(path)
    if path not in sys.path:
        if front:
            sys.path.insert(0, path)
        else:  # Back
            sys.path.append(path)


# Temporarily set the working directory during a block
# Working directory is global, so this can make other threads wait
def set_working_dir(working_dir):
    openc3_chdir_mutex.acquire()
    try:
        current_dir = os.getcwd()
        os.chdir(working_dir)
        yield
    finally:
        openc3_chdir_mutex.release()
        os.chdir(current_dir)


# Attempt to gracefully kill a thread
# @param owner Object that owns the thread and may have a graceful_kill method
# @param thread The thread to gracefully kill
# @param timeout Timeout in seconds to wait for it to die
def kill_thread(owner, thread, timeout=1.0):
    if not thread:
        return

    if owner and hasattr(owner, "graceful_kill"):
        if threading.current_thread() != thread:
            owner.graceful_kill()
            thread.join(timeout=timeout)
        else:
            Logger.warn("Threads cannot graceful_kill themselves")
    elif owner:
        Logger.info(f"Thread owner {owner.__class__.__name__} does not support graceful_kill")
    if thread.is_alive():
        # If the thread dies after alive? but before backtrace, bt will be nil.
        trace = []
        for filename, lineno, name, line in traceback.extract_stack(sys._current_frames()[thread.ident]):
            trace.append(f"{filename}:{lineno}:{name}:{line}")
        caller_trace = []
        for filename, lineno, name, line in traceback.extract_stack(
            sys._current_frames()[threading.current_thread().ident]
        ):
            caller_trace.append(f"{filename}:{lineno}:{name}:{line}")

        # Graceful failed
        caller_trace_string = "\n  ".join(caller_trace)
        trace_string = "\n  ".join(trace)
        msg = "Failed to gracefully kill thread:\n"
        msg = msg + f"  Caller Backtrace:\n  {caller_trace_string}\n"
        msg = msg + f"  \n  Thread Backtrace:\n  {trace_string}\n\n"
        Logger.warn(msg)


# Close a socket in a manner that ensures that any reads blocked in select
# will unblock across platforms
# @param socket The socket to close
def close_socket(socket_to_close):
    if socket_to_close:
        # Calling shutdown and then sleep seems to be required
        # to get select to reliably unblock on linux
        try:
            socket_to_close.shutdown(socket.SHUT_RDWR)
            time.sleep(0)
        except OSError:
            # Oh well we tried
            pass
        try:
            socket_to_close.close()
        # Capture the Socket is not connected error
        except OSError:
            pass


def get_class_from_module(module, class_name):
    """Returns the class from the given module, importing it if necessary"""
    if not sys.modules.get(module):
        parts = module.split(".")
        if len(parts) == 1:
            importlib.import_module(f"{parts[0]}", None)
        else:
            importlib.import_module(f".{parts[-1]}", ".".join(parts[0:-1]))
    return getattr(sys.modules[module], class_name)


# # Import the class represented by the filename. This uses the standard Python
# # convention of having a single class per file where the class name is camel
# # cased and filename is lowercase with underscores.
# #
# # @param class_name_or_class_filename [String] The name of the class or the file which contains the
# #   Python class to import
# # @param log_error [Boolean] Whether to log an error if we can't import the class
# def import_class(class_name_or_class_filename, log_error=True):
#     if class_name_or_class_filename.lower()[-3:] == ".py" or (
#         class_name_or_class_filename[0] == class_name_or_class_filename[0].lower()
#     ):
#         class_filename = class_name_or_class_filename
#         class_name = filename_to_class_name(class_filename)
#     else:
#         class_name = class_name_or_class_filename
#         class_filename = class_name_to_filename(class_name)
#     if to_class(class_name) and sys.modules[class_name]:
#         return to_class(class_name)

#     importlib.import_module(class_filename)
#     klass = to_class(class_name)
#     if klass is None:
#         raise RuntimeError(f"Python class {class_name} not found")

#     return klass



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import math
import json
import struct
from datetime import datetime, timezone
from openc3.config.config_parser import ConfigParser
from openc3.interfaces.protocols.burst_protocol import BurstProtocol


# Delineates packets using the OpenC3 preidentification system
class PreidentifiedProtocol(BurstProtocol):
    COSMOS4_STORED_FLAG_MASK = 0x80
    COSMOS4_EXTRA_FLAG_MASK = 0x40

    # @param sync_pattern (see BurstProtocol#initialize)
    # @param max_length [Integer] The maximum allowed value of the length field
    # @param _unused [Integer] Legacy version number - unused
    # @param allow_empty_data [true/false/nil] See Protocol#initialize
    def __init__(self, sync_pattern=None, max_length=None, _unused=None, allow_empty_data=None):
        super().__init__(0, sync_pattern, False, allow_empty_data)
        self.max_length = ConfigParser.handle_none(max_length)
        if self.max_length:
            self.max_length = int(self.max_length)

    def reset(self):
        super().reset()
        self.reduction_state = "START"

    def read_packet(self, packet):
        packet.received_time = self.read_received_time
        packet.target_name = self.read_target_name.decode(encoding="ascii")
        packet.packet_name = self.read_packet_name.decode(encoding="ascii")
        packet.stored = self.read_stored
        if packet.extra and self.read_extra:
            packet.extra.merge(self.read_extra)
        else:
            packet.extra = self.read_extra
        return packet

    def write_packet(self, packet):
        received_time = packet.received_time
        if not received_time:
            received_time = datetime.now(timezone.utc)
        tv_usec, tv_sec = math.modf(received_time.timestamp())
        self.write_time_seconds = struct.pack(">I", int(tv_sec))  # UINT32
        self.write_time_microseconds = struct.pack(">I", int(tv_usec * 1_000_000))  # UINT32
        self.write_target_name = packet.target_name
        if not self.write_target_name:
            self.write_target_name = "UNKNOWN"
        self.write_packet_name = packet.packet_name
        if not self.write_packet_name:
            self.write_packet_name = "UNKNOWN"
        self.write_flags = 0
        if packet.stored:
            self.write_flags |= PreidentifiedProtocol.COSMOS4_STORED_FLAG_MASK
        self.write_extra = None
        if packet.extra:
            self.write_flags |= PreidentifiedProtocol.COSMOS4_EXTRA_FLAG_MASK
            self.write_extra = json.dumps(packet.extra)
        return packet

    def write_data(self, data, extra=None):
        data_length = struct.pack(">I", len(data))  # UINT32
        data_to_send = b""
        if self.sync_pattern:
            data_to_send += self.sync_pattern
        data_to_send += self.write_flags.to_bytes(1, byteorder="big")
        if self.write_extra:
            data_to_send += struct.pack(">I", len(self.write_extra))
            data_to_send += bytes(self.write_extra, "ascii")
        data_to_send += self.write_time_seconds
        data_to_send += self.write_time_microseconds
        data_to_send += struct.pack(">B", len(self.write_target_name))
        data_to_send += bytes(self.write_target_name, "ascii")
        data_to_send += struct.pack(">B", len(self.write_packet_name))
        data_to_send += bytes(self.write_packet_name, "ascii")
        data_to_send += data_length
        data_to_send += data
        return (data_to_send, extra)

    def read_length_field_followed_by_string(self, length_num_bytes):
        # Read bytes for string length
        if len(self.data) < length_num_bytes:
            return "STOP"

        string_length = self.data[0:length_num_bytes]

        match length_num_bytes:
            case 1:
                string_length = struct.unpack(">B", string_length)[0]  # UINT8
            case 2:
                string_length = struct.unpack(">H", string_length)[0]  # UINT16
            case 4:
                string_length = struct.unpack(">I", string_length)[0]  # UINT32
                if self.max_length and string_length > self.max_length:
                    raise RuntimeError(
                        f"Length value received larger than max_length: {string_length} > {self.max_length}"
                    )
            case _:
                raise RuntimeError(
                    f"Unsupported length given to read_length_field_followed_by_string: {length_num_bytes}"
                )

        # Read String
        if len(self.data) < (string_length + length_num_bytes):
            return "STOP"

        next_index = string_length + length_num_bytes
        string = self.data[length_num_bytes:next_index]

        # Remove data from current_data
        self.data = self.data[next_index:]

        return string

    def reduce_to_single_packet(self):
        # Discard sync pattern if present
        if self.sync_pattern:
            if self.reduction_state == "START":
                if len(self.data) < len(self.sync_pattern):
                    return ("STOP", self.extra)

                self.data = self.data[(len(self.sync_pattern)) :]
                self.reduction_state = "SYNC_REMOVED"
        elif self.reduction_state == "START":
            self.reduction_state = "SYNC_REMOVED"

        if self.reduction_state == "SYNC_REMOVED":
            # Read and remove flags
            if len(self.data) < 1:
                return ("STOP", self.extra)

            flags = self.data[0]  # struct.unpack(">B", self.data[0])  # byte
            self.data = self.data[1:]
            self.read_stored = False
            if (flags & PreidentifiedProtocol.COSMOS4_STORED_FLAG_MASK) != 0:
                self.read_stored = True
            self.read_extra = None
            if (flags & PreidentifiedProtocol.COSMOS4_EXTRA_FLAG_MASK) != 0:
                self.reduction_state = "NEED_EXTRA"
            else:
                self.reduction_state = "FLAGS_REMOVED"

        if self.reduction_state == "NEED_EXTRA":
            # Read and remove extra
            self.read_extra = self.read_length_field_followed_by_string(4)
            if self.read_extra == "STOP":
                return ("STOP", self.extra)

            self.read_extra = json.loads(self.read_extra)
            self.reduction_state = "FLAGS_REMOVED"

        if self.reduction_state == "FLAGS_REMOVED":
            # Read and remove packet received time
            if len(self.data) < 8:
                return ("STOP", self.extra)

            time_seconds = struct.unpack(">I", self.data[0:4])[0]  # UINT32
            time_microseconds = struct.unpack(">I", self.data[4:8])[0]  # UINT32
            self.read_received_time = datetime.fromtimestamp(time_seconds + time_microseconds / 1_000_000, timezone.utc)
            self.data = self.data[8:]
            self.reduction_state = "TIME_REMOVED"

        if self.reduction_state == "TIME_REMOVED":
            # Read and remove the target name
            self.read_target_name = self.read_length_field_followed_by_string(1)
            if self.read_target_name == "STOP":
                return ("STOP", self.extra)

            self.reduction_state = "TARGET_NAME_REMOVED"

        if self.reduction_state == "TARGET_NAME_REMOVED":
            # Read and remove the packet name
            self.read_packet_name = self.read_length_field_followed_by_string(1)
            if self.read_packet_name == "STOP":
                return ("STOP", self.extra)

            self.reduction_state = "PACKET_NAME_REMOVED"

        if self.reduction_state == "PACKET_NAME_REMOVED":
            # Read packet data and return
            packet_data = self.read_length_field_followed_by_string(4)
            if packet_data == "STOP":
                return ("STOP", self.extra)

            self.reduction_state = "START"
            return (packet_data, self.extra)

        raise RuntimeError(f"Error should never reach end of method {self.reduction_state}")



# Copyright 2023 OpenC3, Inc
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.:

from openc3.api import WHITELIST
from openc3.environment import OPENC3_SCOPE
from openc3.utilities.authorization import authorize
from openc3.models.target_model import TargetModel
from openc3.models.interface_model import InterfaceModel

WHITELIST.extend(
    [
        "get_target_names",
        "get_target",
        "get_target_interfaces",
    ]
)


# Returns the list of all target names
#
# @return [Array<String>] All target names
def get_target_names(scope=OPENC3_SCOPE):
    authorize(permission="system", scope=scope)
    return TargetModel.names(scope=scope)


# Gets the full target hash
#
# @since 5.0.0
# @param target_name [String] Target name
# @return [Hash] Hash of all the target properties
def get_target(target_name, scope=OPENC3_SCOPE):
    authorize(permission="system", target_name=target_name, scope=scope)
    return TargetModel.get(name=target_name, scope=scope)


# Get all targets and their interfaces
#
# @return [Array<Array<String, String] Array of Arrays \[name, interfaces]
def get_target_interfaces(scope=OPENC3_SCOPE):
    authorize(permission="system", scope=scope)
    info = []
    interfaces = InterfaceModel.all(scope=scope)
    for target_name in get_target_names(scope=scope):
        interface_names = []
        for _, interface in interfaces.items():
            if target_name in interface["target_names"]:
                interface_names.append(interface["name"])
        info.append([target_name, ",".join(interface_names)])
    return info



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import threading
import time
import json
from openc3.environment import *
from requests import Session


# Basic exception for known errors
class OpenC3AuthenticationError(RuntimeError):
    pass


class OpenC3AuthenticationRetryableError(OpenC3AuthenticationError):
    pass


# OpenC3 COSMOS Core authentication code
class OpenC3Authentication:
    def __init__(self):
        self._token = OPENC3_API_PASSWORD
        if not self._token:
            raise OpenC3AuthenticationError("Authentication requires environment variable OPENC3_API_PASSWORD")

    def token(self, include_bearer=True):
        return self._token


# OpenC3 enterprise Keycloak authentication code
class OpenC3KeycloakAuthentication(OpenC3Authentication):
    # {
    #     "access_token": "",
    #     "expires_in": 600,
    #     "refresh_expires_in": 1800,
    #     "refresh_token": "",
    #     "token_type": "bearer",
    #     "id_token": "",
    #     "not-before-policy": 0,
    #     "session_state": "",
    #     "scope": "openid email profile"
    # }

    REFRESH_OFFSET_SECONDS = 60

    # @param url [String] The url of the openc3 or keycloak in the cluster
    def __init__(self, url):
        self.url = url
        self.auth_mutex = threading.Lock()
        self.refresh_token = None
        self.expires_at = None
        self.refresh_expires_at = None
        self._token = None
        self.log = [None, None]
        self.http = Session()

    # Load the token from the environment
    def token(self, include_bearer=True, openid_scope="openid"):
        with self.auth_mutex:
            self.log = [None, None]
            current_time = time.time()
            if self._token is None:
                self._make_token(current_time, openid_scope=openid_scope)
            elif self.refresh_expires_at < current_time:
                self._make_token(current_time, openid_scope=openid_scope)
            elif self.expires_at < current_time:
                self._refresh_token(current_time)
        if include_bearer:
            return f"Bearer {self._token}"
        else:
            return self._token

    def get_token_from_refresh_token(self, refresh_token):
        current_time = time.time()
        try:
            self.refresh_token = refresh_token
            self._refresh_token(current_time)
            return self._token
        except OpenC3AuthenticationError:
            return None

    # Make the token and save token to instance
    def _make_token(self, current_time, openid_scope="openid"):
        client_id = OPENC3_API_CLIENT or "api"
        if OPENC3_API_USER and OPENC3_API_PASSWORD:
            # Username and password
            data = f"username={OPENC3_API_USER}&password={OPENC3_API_PASSWORD}&client_id={client_id}&grant_type=password&scope={openid_scope}"
            headers = {
                "Content-Type": "application/x-www-form-urlencoded",
                "User-Agent": OPENC3_USER_AGENT,
            }
            oath = self._make_request(headers, data)
            self._token = oath["access_token"]
            self.refresh_token = oath["refresh_token"]
            self.expires_at = current_time + oath["expires_in"] - self.REFRESH_OFFSET_SECONDS
            self.refresh_expires_at = current_time + oath["refresh_expires_in"] - self.REFRESH_OFFSET_SECONDS
        else:
            # Offline Access Token
            if self.refresh_token is None:
                self.refresh_token = OPENC3_API_TOKEN
            self._refresh_token(current_time)
        return None

    # Refresh the token and save token to instance
    def _refresh_token(self, current_time):
        client_id = OPENC3_API_CLIENT or "api"
        data = f"client_id={client_id}&refresh_token={self.refresh_token}&grant_type=refresh_token"
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "User-Agent": OPENC3_USER_AGENT,
        }
        oath = self._make_request(headers, data)
        self._token = oath["access_token"]
        self.refresh_token = oath["refresh_token"]
        self.expires_at = current_time + oath["expires_in"] - self.REFRESH_OFFSET_SECONDS
        self.refresh_expires_at = current_time + oath["refresh_expires_in"] - self.REFRESH_OFFSET_SECONDS

    # Make the post request to keycloak
    def _make_request(self, headers, data):
        realm = OPENC3_KEYCLOAK_REALM or "openc3"
        url = f"{self.url}/realms/{realm}/protocol/openid-connect/token"
        request_kwargs = {
            "url": url,
            "data": data,
            "headers": headers,
        }
        self.log[0] = f"Request: {request_kwargs}"
        # print(self.log[0])
        resp = self.http.post(**request_kwargs)
        self.log[1] = f"response status: {resp.status_code} header: {resp.headers} body: {resp.text}"
        # print(self.log[1])
        if resp.status_code >= 200 and resp.status_code <= 299:
            return json.loads(resp.text)
        elif resp.status_code >= 500 and resp.status_code <= 599:
            raise OpenC3AuthenticationRetryableError(
                f"authentication request retryable {self.log[0]} ::: {self.log[1]}"
            )
        else:
            raise OpenC3AuthenticationError(f"authentication request failed {self.log[0]} ::: {self.log[1]}")



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import socket
import select
import ipaddress


class UdpReadWriteSocket:
    # @param bind_port [Integer[ Port to write data out from and receive data on (0 = randomly assigned)
    # @param bind_address [String] Local address to bind to (0.0.0.0 = All local addresses)
    # @param external_port [Integer] External port to write to
    # @param external_address [String] External host to send data to
    # @param multicast_interface_address [String] Local outgoing interface to send multicast packets from
    # @param ttl [Integer] Time To Live for outgoing multicast packets
    # @param read_multicast [Boolean] Whether or not to try to read from the external address as multicast
    # @param write_multicast [Boolean] Whether or not to write to the external address as multicast
    def __init__(
        self,
        bind_port=0,
        bind_address="0.0.0.0",
        external_port=None,
        external_address=None,
        multicast_interface_address=None,
        ttl=1,
        read_multicast=True,
        write_multicast=True,
    ):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        # Basic setup to reuse address
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        # Bind to local address and port - This sets recv port, write_src port, recv_address, and write_src_address
        if bind_address and bind_port:
            self.socket.bind((bind_address, bind_port))

        # Default send to the specified address and port
        if external_address and external_port:
            self.socket.connect((external_address, external_port))

        # Handle multicast
        if UdpReadWriteSocket.multicast(external_address, external_port):
            if write_multicast:
                # Basic setup set time to live
                self.socket.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, int(ttl))

                if multicast_interface_address:
                    # Set outgoing interface
                    self.socket.setsockopt(
                        socket.SOL_IP,
                        socket.IP_MULTICAST_IF,
                        socket.inet_aton(multicast_interface_address),
                    )

            # Receive messages sent to the multicast address
            if read_multicast:
                if not multicast_interface_address:
                    multicast_interface_address = "0.0.0.0"
                membership = socket.inet_aton(external_address) + socket.inet_aton(multicast_interface_address)
                self.socket.setsockopt(socket.SOL_IP, socket.IP_ADD_MEMBERSHIP, membership)

    # @param data [String] Binary data to send
    # @param write_timeout [Float] Time in seconds to wait for the data to send
    def write(self, data, write_timeout=10.0):
        num_bytes_to_send = len(data)
        total_bytes_sent = 0
        bytes_sent = 0
        data_to_send = data

        while True:
            try:
                bytes_sent = self.socket.send(data_to_send)
            except socket.error as e:
                if e.args[0] == socket.EAGAIN or e.args[0] == socket.EWOULDBLOCK:
                    result = select.select([], [self.socket], [], write_timeout)
                    if len(result[0]) == 0 and len(result[1]) == 0 and len(result[2]) == 0:
                        raise socket.timeout
            total_bytes_sent += bytes_sent
            if total_bytes_sent >= num_bytes_to_send:
                break
        data_to_send = data[total_bytes_sent:]
        return data_to_send

    # @param read_timeout [Float] Time in seconds to wait for the read to
    #   complete
    def read(self, read_timeout=None):
        data = None
        while True:
            try:
                data, _ = self.socket.recvfrom(65536, socket.MSG_DONTWAIT)
            except socket.error as e:
                if e.args[0] == socket.EAGAIN or e.args[0] == socket.EWOULDBLOCK:
                    result = select.select([self.socket], [], [], read_timeout)
                    if len(result[0]) == 0 and len(result[1]) == 0 and len(result[2]) == 0:
                        raise socket.timeout
                    else:
                        continue
            break
        return data

    # Defer all methods to the UDPSocket
    def __getattr__(self, func):
        def method(*args, **kwargs):
            return getattr(self.socket, func)(*args, **kwargs)

        return method

    # @param host [String] Machine name or IP address
    # @param port [String] Port
    # @return [Boolean] Whether the hostname is multicast
    @classmethod
    def multicast(cls, host, port):
        if host is None or port is None:
            return False
        host_ip = socket.gethostbyname(host)
        return ipaddress.ip_address(host_ip) in ipaddress.ip_network("224.0.0.0/4")


# Creates a UDPSocket and implements a non-blocking write.
class UdpWriteSocket(UdpReadWriteSocket):
    # @param dest_address [String] Host to send data to
    # @param dest_port [Integer] Port to send data to
    # @param src_port [Integer[ Port to send data out from
    # @param multicast_interface_address [String] Local outgoing interface to send multicast packets from
    # @param ttl [Integer] Time To Live for outgoing packets
    # @param bind_address [String] Local address to bind to (0.0.0.0 = All local addresses)
    def __init__(
        self,
        dest_address,
        dest_port,
        src_port=None,
        multicast_interface_address=None,
        ttl=1,
        bind_address="0.0.0.0",
    ):
        super().__init__(
            src_port,
            bind_address,
            dest_port,
            dest_address,
            multicast_interface_address,
            ttl,
            False,
            True,
        )


# Creates a UDPSocket and implements a non-blocking read.
class UdpReadSocket(UdpReadWriteSocket):
    # @param recv_port [Integer] Port to receive data on
    # @param multicast_address [String] Address to add multicast
    # @param multicast_interface_address [String] Local incoming interface to receive multicast packets on
    # @param bind_address [String] Local address to bind to (0.0.0.0 = All local addresses)
    def __init__(
        self,
        recv_port=0,
        multicast_address=None,
        multicast_interface_address=None,
        bind_address="0.0.0.0",
    ):
        super().__init__(
            recv_port,
            bind_address,
            None,
            multicast_address,
            multicast_interface_address,
            1,
            True,
            False,
        )



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.interfaces.interface import Interface
from openc3.config.config_parser import ConfigParser
from openc3.utilities.logger import Logger
from openc3.utilities.string import class_name_to_filename
from openc3.top_level import get_class_from_module


# Base class for interfaces that act read and write from a stream
class StreamInterface(Interface):
    def __init__(self, protocol_type=None, protocol_args=[]):
        super().__init__()
        self.stream = None
        self.protocol_type = ConfigParser.handle_none(protocol_type)
        self.protocol_args = protocol_args
        if self.protocol_type:
            protocol_class_name = str(protocol_type).capitalize() + "Protocol"
            filename = class_name_to_filename(protocol_class_name)
            klass = get_class_from_module(f"openc3.interfaces.protocols.{filename}", protocol_class_name)
            self.add_protocol(klass, protocol_args, "PARAMS")

    def connect(self):
        super().connect() # Reset the protocols
        if self.stream:
            self.stream.connect()

    def connected(self):
        if self.stream:
            return self.stream.connected()
        else:
            return False

    def disconnect(self):
        if self.stream:
            self.stream.disconnect()
        super().disconnect()

    def read_interface(self):
        timeout = False
        try:
            data = self.stream.read()
        except TimeoutError:
            Logger.error(f"{self.name}: Timeout waiting for data to be read")
            timeout = True
            data = None
        if data is None or len(data) <= 0:
            if data is None and not timeout:
                Logger.info(f"{self.name}: {self.stream.__class__.__name__} read returned None")
            if data is not None and len(data) <= 0:
                Logger.info(f"{self.name}: {self.stream.__class__.__name__} read returned 0 bytes (stream closed)")
            return None, None

        extra = None
        self.read_interface_base(data, extra)
        return data, extra

    def write_interface(self, data, extra=None):
        self.write_interface_base(data, extra)
        self.stream.write(data)
        return data, extra



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import json
import time
from openc3.topics.topic import Topic
from openc3.environment import OPENC3_SCOPE
from openc3.utilities.json import JsonEncoder, JsonDecoder


class DecomInterfaceTopic(Topic):
    @classmethod
    def build_cmd(cls, target_name, cmd_name, cmd_params, range_check, raw, timeout=5, scope=OPENC3_SCOPE):
        data = {}
        data["target_name"] = target_name.upper()
        data["cmd_name"] = cmd_name.upper()
        data["cmd_params"] = cmd_params
        data["range_check"] = range_check
        data["raw"] = raw
        # DecomMicroservice is listening to the DECOMINTERFACE topic and is responsible
        # for actually building the command. This was deliberate to allow this to work
        # with or without an interface.
        ack_topic = f"{{{scope}__ACKCMD}}TARGET__{target_name}"
        Topic.update_topic_offsets([ack_topic])
        decom_id = Topic.write_topic(
            f"{scope}__DECOMINTERFACE__{{{target_name}}}",
            {"build_cmd": json.dumps(data, cls=JsonEncoder)},
            "*",
            100,
        )
        start_time = time.time()
        while (time.time() - start_time) < timeout:
            for _topic, _msg_id, msg_hash, _redis in Topic.read_topics([ack_topic]):
                if msg_hash[b"id"] == decom_id:
                    if msg_hash[b"result"] == b"SUCCESS":
                        msg_hash = {k.decode(): v.decode() for (k, v) in msg_hash.items()}
                        msg_hash["buffer"] = json.loads(msg_hash["buffer"], cls=JsonDecoder)
                        return msg_hash
                    else:
                        raise RuntimeError(msg_hash[b"message"])
        raise RuntimeError(f"Timeout of {timeout}s waiting for cmd ack. Does target '{target_name}' exist?")

    @classmethod
    def inject_tlm(
        cls,
        target_name,
        packet_name,
        item_hash=None,
        type="CONVERTED",
        scope=OPENC3_SCOPE,
    ):
        data = {}
        data["target_name"] = target_name.upper()
        data["packet_name"] = packet_name.upper()
        data["item_hash"] = item_hash
        data["type"] = type
        Topic.write_topic(
            f"{scope}__DECOMINTERFACE__{{{target_name}}}",
            {"inject_tlm": json.dumps(data)},
            "*",
            100,
        )



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import os
import sys
import re
from openc3.utilities.extract import remove_quotes


class ConfigParser:
    """Reads OpenC3 style configuration data which consists of keywords followed
    by 0 or more comma delimited parameters. Parameters with spaces must be
    enclosed in quotes. Quotes should also be used to indicate a parameter is a
    string. Keywords are case-insensitive and will be returned in uppercase."""

    # Regular expression used to break up an individual line into a keyword and
    # comma delimited parameters. Handles parameters in single or double quotes.
    PARSING_REGEX = r"(?:\"(?:[^\\\"]|\\.)*\") | (?:'(?:[^\\']|\\.)*') | \S+"

    class Error(Exception):
        """Error which gets raised by ConfigParser in #verify_num_parameters. This
        is also the error that classes using ConfigParser should raise when they
        encounter a configuration error."""

        # self.param config_parser [ConfigParser] Instance of ConfigParser so Error
        #   has access to the ConfigParser attributes
        # self.param message [String] The error message which gets passed to the
        #   StandardError constructor
        # self.param usage [String] The usage string representing how this keyword should
        #   be formatted.
        # self.param url [String] URL which should point to usage information. By
        #   default this gets constructed to point to the generic configuration
        #   Guide on the OpenC3 Wiki.
        def __init__(self, config_parser, message="Configuration Error", usage="", url=""):
            super().__init__(message)
            self.keyword = config_parser.keyword
            self.parameters = config_parser.parameters
            self.filename = config_parser.filename
            self.line = config_parser.line
            self.line_number = config_parser.line_number
            self.usage = usage
            self.url = url

    # self.param url [String] The url to link to in error messages
    def __init__(self, url="https://docs.openc3.com/docs"):
        self.url = url
        self.keyword = None
        self.parameters = None
        self.filename = None
        self.line = None
        self.line_number = None

    # self.param message [String] The string to set the Exception message to
    # self.param usage [String] The usage message
    # self.param url [String] Where to get help about this error
    # self.return [Error] The constructed error
    def error(self, message, usage="", url=None):
        if not url:
            url = self.url
        return self.Error(self, message, usage, url)

    # Can be called during parsing to read a referenced file
    def read_file(self, filename):
        # Assume the file is there. If not we raise a pretty obvious error
        if os.path.abspath(filename) == filename:  # absolute path
            path = filename
        else:  # relative to the current @filename
            path = os.path.join(os.path.dirname(self.filename), filename)
        data = ""
        with open(path, "rb") as file:
            data = file.read()
        return data

    def parse_file(
        self,
        filename,
        yield_non_keyword_lines=False,
        remove_quotes_arg=True,
        run_erb=True,
        variables={}
    ):
        """Parse a file and call the callback for each line

        Args:
            filename (str): The file to parse
            yield_non_keyword_lines (bool, optional): Whether to yield lines without keywords
            remove_quotes_arg (bool, optional): Whether to remove quotes from parameters
            run_erb (bool, optional): No effect in Python
            variables (dict, optional): No effect in Python

        Returns:
            None
        """
        if filename and not os.path.exists(filename):
            raise ConfigParser.Error(self, f"Configuration file {filename} does not exist.")

        self.filename = filename
        with open(filename, "r") as file:
            # Loop through each line of the data
            yield from self.parse_loop(
                file,
                yield_non_keyword_lines,
                remove_quotes_arg,
                os.path.getsize(file.name),
                ConfigParser.PARSING_REGEX,
            )

    def verify_num_parameters(self, min_num_params, max_num_params, usage=""):
        """
        Verifies the parameters in the ConfigParser have the specified number of arguments

        Args:
            min_num_params (int): The minimum number of parameters
            max_num_params (int): The maximum number of parameters
            usage (str): Usage description to display in error message

        Raises:
            Error: Insufficient/Excessive parameters for keyword
        """
        # This syntax works with 0 because each doesn't return any values
        # for a backwards range
        for index in range(1, min_num_params + 1):
            # If the parameter is None (0 based) then we have a problem
            if not self.parameters[index - 1 : index]:
                raise ConfigParser.Error(self, f"Not enough parameters for {self.keyword}.", usage, self.url)

        # If they pass None for max_params we don't check for a maximum number
        if max_num_params is not None and self.parameters[max_num_params : max_num_params + 1]:
            raise ConfigParser.Error(self, f"Too many parameters for {self.keyword}.", usage, self.url)

    def verify_parameter_naming(self, index, usage=""):
        """
        Verifies that the parameter does not contain reserved characters

        Args:
            index (int): The index of the parameter to be verified
        """
        param = self.parameters[index - 1]
        if param[-1] == "_":
            raise ConfigParser.Error(
                self,
                f"Parameter {index} ({param}) for {self.keyword} cannot end with an underscore ('_').",
                usage,
                self.url,
            )

        if "__" in param:
            raise ConfigParser.Error(
                self,
                f"Parameter {index} ({param}) for {self.keyword} cannot contain a double underscore ('__').",
                usage,
                self.url,
            )

        if "[[" in param or "]]" in param:
            raise ConfigParser.Error(
                self,
                f"Parameter {index} ({param}) for {self.keyword} cannot contain double brackets ('[[' or ']]').",
                usage,
                self.url,
            )

        if " " in param:
            raise ConfigParser.Error(
                self,
                f"Parameter {index} ({param}) for {self.keyword} cannot contain a space (' ').",
                usage,
                self.url,
            )

        if "'" in param or '"' in param:
            raise ConfigParser.Error(
                self,
                f"Parameter {index} ({param}) for {self.keyword} cannot contain a quote (' or \").",
                usage,
                self.url,
            )

        if "{" in param or "}" in param:
            raise ConfigParser.Error(
                self,
                f"Parameter {index} ({param}) for {self.keyword} cannot contain a curly bracket ('{{' or '}}').",
                usage,
                self.url,
            )

    # Converts a String containing '', 'NONE', or 'NULL' to Python primitive.
    # All other arguments are simply returned.
    #
    # self.param value [Object]
    # self.return [None|Object]
    @classmethod
    def handle_none(cls, value):
        if isinstance(value, str):
            match value.upper():
                case "" | "NONE" | "NULL":
                    return None

        return value

    # Converts a String containing 'TRUE' or 'FALSE' to True or False Python
    # primitive. All other values are simply returned.
    #
    # self.param value [Object]
    # self.return [True|False|Object]
    @classmethod
    def handle_true_false(cls, value):
        if isinstance(value, str):
            match value.upper():
                case "TRUE":
                    return True
                case "FALSE":
                    return False
        return value

    # Converts a String containing '', 'NONE', 'NULL', 'NIL', 'TRUE' or 'FALSE' to None,
    # True or False Python primitives. All other values are simply returned.
    #
    # self.param value [Object]
    # self.return [True|False|None|Object]
    @classmethod
    def handle_true_false_none(cls, value):
        if isinstance(value, str):
            match value.upper():
                case "TRUE":
                    return True
                case "FALSE":
                    return False
                # Convert nil for the Rubyists
                case "" | "NONE" | "NULL" | "NIL":
                    return None
        return value

    # Converts a string representing a defined constant into its value. The
    # defined constants are the minimum and maximum values for all the
    # allowable data types. [MIN/MAX]_[U]INT[8/16/32] and
    # [MIN/MAX]_FLOAT[32/64]. Thus MIN_UINT8, MAX_INT32, and MIN_FLOAT64 are
    # all allowable values. Any other strings raise ArgumentError but all other
    # types are simply returned.
    #
    # self.param value [Object] Can be anything
    # self.return [Numeric] The converted value. Either a Fixnum or Float.
    @classmethod
    def handle_defined_constants(cls, value, data_type=None, bit_size=None):
        if isinstance(value, str):
            match value.upper():
                case "MIN" | "MAX":
                    return ConfigParser.calculate_range_value(value.upper(), data_type, bit_size)
                case "MIN_INT8":
                    return -128
                case "MAX_INT8":
                    return 127
                case "MIN_INT16":
                    return -32768
                case "MAX_INT16":
                    return 32767
                case "MIN_INT32":
                    return -2147483648
                case "MAX_INT32":
                    return 2147483647
                case "MIN_INT64":
                    return -9223372036854775808
                case "MAX_INT64":
                    return 9223372036854775807
                case "MIN_UINT8" | "MIN_UINT16" | "MIN_UINT32" | "MIN_UINT64":
                    return 0
                case "MAX_UINT8":
                    return 255
                case "MAX_UINT16":
                    return 65535
                case "MAX_UINT32":
                    return 4294967295
                case "MAX_UINT64":
                    return 18446744073709551615
                case "MIN_FLOAT64":
                    return -sys.float_info.max
                case "MAX_FLOAT64":
                    return sys.float_info.max
                case "MIN_FLOAT32":
                    return -3.402823e38
                case "MAX_FLOAT32":
                    return 3.402823e38
                case "POS_INFINITY":
                    return float("inf")
                case "NEG_INFINITY":
                    return float("-inf")
                # NOTE: No else case because of the following scenario:
                # If the value type is a UINT but they have a WRITE_CONVERSION that takes a string
                # then the default value will be a string. In that case we just want to return the string.
                # For example, the IP_ADDRESS parameter in the TIME_OFFSET command in the Demo plugin.

        return value

    @classmethod
    def calculate_range_value(cls, type, data_type, bit_size):
        """
        Calculate the min or max value for a given data type and bit size

        Args:
            data_type: Data type (INT, UINT, etc.)
            bit_size: Size in bits
            type: MIN or MAX

        Returns:
            Min or max value
        """
        value = 0  # Default for UINT minimum

        match data_type:
            case "INT":
                if type == "MIN":
                    value = -(2 ** (bit_size - 1))
                else:  # 'MAX'
                    value = 2 ** (bit_size - 1) - 1

            case "UINT":
                # Default is 0 for 'MIN'
                if type == "MAX":
                    value = 2**bit_size - 1

            case "FLOAT":
                match bit_size:
                    case 32:
                        value = 3.402823e38
                        if type == "MIN":
                            value *= -1
                    case 64:
                        value = sys.float_info.max
                        if type == "MIN":
                            value *= -1
                    case _:
                        raise ValueError(f"Invalid bit size {bit_size} for FLOAT type.")

            case _:
                raise TypeError(f"Invalid data type {data_type} when calculating range.")

        return value

    def parse_errors(self, errors):
        if len(errors) == 0:
            return
        message = ""
        for error in errors:
            if issubclass(error, ConfigParser.Error):
                message += f"\n{os.path.basename(error.filename)}:{error.line_number}: {error.line}"
                message += f"\nError: {repr(error)}"
                if not error.usage:
                    message += f"\nUsage: {error.usage}"
            else:
                message += f"\n{repr(error)}"
        message += "\n"
        raise ConfigParser.Error(self, message)

    # Iterates over each line of the io object and yields the keyword and parameters
    def parse_loop(self, io, yield_non_keyword_lines, remove_quotes_arg, size, rx):
        string_concat = False
        self.line_number = 0
        self.keyword = None
        self.parameters = []
        self.line = ""
        errors = []

        while line := io.readline():
            self.line_number += 1

            line = line.strip()
            # Ensure the line length is not 0
            if len(line) == 0:
                continue

            if string_concat:
                # Skip comment lines after a string concatenation
                if line[0] == "#":
                    continue
                # Remove the opening quote if we're continuing the line
                line = line[1:]

            # Check for string continuation
            match line[-1]:
                case "+" | "\\":  # String concatenation
                    newline = line[-1] == "+"
                    # Trim off the concat character plus any spaces, e.g. "line" \
                    trim = line[0:-1].strip()
                    # Now trim off the last quote so it will flow into the next line
                    self.line += trim[0:-1]
                    if newline:
                        self.line += "\n"
                    string_concat = True
                    continue
                case "&":  # Line continuation
                    self.line += line[0:-1]
                    continue
                case _:
                    self.line += line
            string_concat = False

            data = re.compile(rx, re.X).findall(self.line)
            first_item = ""
            if len(data) > 0:
                first_item += data[0]

            if (len(first_item) == 0) or (first_item[0] == "#"):
                self.keyword = None
            else:
                self.keyword = first_item.upper()
            self.parameters = []

            # Ignore lines without keywords: comments and blank lines
            if not self.keyword:
                if yield_non_keyword_lines:
                    try:
                        yield self.keyword, self.parameters
                    except Exception as error:
                        errors.append(error)
                self.line = ""
                continue

            length = len(data)
            if length > 1:
                for index in range(1, length):
                    string = data[index]

                    # Don't process trailing comments such as:
                    # KEYWORD PARAM #This is a comment
                    # But still process Ruby string interpolations such as:
                    # KEYWORD PARAM {var}
                    if (len(string) > 0) and (string[0] == "#"):
                        if not ((len(string) > 1) and (string[1] == "{")):
                            break

                    if remove_quotes_arg:
                        self.parameters.append(remove_quotes(string))
                    else:
                        self.parameters.append(string)

            try:
                yield self.keyword, self.parameters
            except Exception as error:
                errors.append(error)
            self.line = ""

        self.parse_errors(errors)
        return None



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from .accessor import Accessor
import urllib.parse


class FormAccessor(Accessor):
    @classmethod
    def read_item(cls, item, buffer):
        ary = urllib.parse.parse_qsl(buffer)
        value = None
        for key, ary_value in ary:
            if key.decode() == item.key:
                # Handle the case of multiple values for the same key
                # and build up an array of values
                if value:
                    # Second time through value is not a list yet
                    if type(value) is not list:
                        value = [value]
                    value.append(ary_value)
                else:
                    value = ary_value
        return value

    @classmethod
    def write_item(cls, item, value, buffer):
        ary = urllib.parse.parse_qsl(buffer)
        # Remove existing item and bad keys from list
        ary = [ary_value for ary_value in ary if (ary_value[0] != item.key) and (str(ary_value[0])[0] != "\u0000")]

        if isinstance(value, list):
            for value_value in value:
                ary.append((item.key, value_value))
        else:
            ary.append((item.key, value))
        buffer[:] = urllib.parse.urlencode(ary).encode()
        return value

    # If this is set it will enforce that buffer data is encoded
    # in a specific encoding
    def enforce_encoding(self):
        return None

    # This affects whether the Packet class enforces the buffer
    # length at all.  Set to false to remove any correlation between
    # buffer length and defined sizes of items in COSMOS
    def enforce_length(self):
        return False

    # This sets the short_buffer_allowed flag in the Packet class
    # which allows packets that have a buffer shorter than the defined size.
    # Note that the buffer is still resized to the defined length
    def enforce_short_buffer_allowed(self):
        return True

    # If this is true it will enforce that COSMOS DERIVED items must have a
    # write_conversion to be written
    def enforce_derived_write_conversion(self, _item):
        return True



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.


import threading
import schedule
import traceback
import time
from contextlib import contextmanager
from datetime import datetime, timezone
from openc3.api import *
from openc3.utilities.logger import Logger
from openc3.utilities.secrets import Secrets
from openc3.logs.stream_log_pair import StreamLogPair


class WriteRejectError(RuntimeError):
    """Define a class to allow interfaces and protocols to reject commands without disconnecting the interface"""

    pass


class Interface:
    """Defines all the attributes and methods common to all interface classes used by OpenC3."""

    # Initialize default attribute values
    def __init__(self):
        self.state = "DISCONNECTED"
        self.target_names = []
        self.cmd_target_names = []
        self.tlm_target_names = []
        self.connect_on_startup = True
        self.auto_reconnect = True
        self.reconnect_delay = 5.0
        self.disable_disconnect = False
        self.packet_log_writer_pairs = []
        self.stored_packet_log_writer_pairs = []
        self.routers = []
        self.cmd_routers = []
        self.read_count = 0
        self.write_count = 0
        self.bytes_read = 0
        self.bytes_written = 0
        self._num_clients = 0
        self._read_queue_size = 0
        self._write_queue_size = 0
        self.write_mutex = threading.RLock()
        self.read_allowed = True
        self.write_allowed = True
        self.write_raw_allowed = True
        self.options = {}
        self.read_protocols = []
        self.write_protocols = []
        self.protocol_info = []
        self.read_raw_data = ""
        self.written_raw_data = ""
        self.read_raw_data_time = None
        self.written_raw_data_time = None
        self.config_params = []
        self.interfaces = []
        self.stream_log_pair = None
        self.secrets = Secrets.getClient()
        self.name = self.__class__.__name__
        self.scheduler = None
        self.scheduler_thread = None
        self.cancel_scheduler_thread = False

    # @return [Integer] The number of connected clients
    def num_clients(self):
        return self._num_clients

    # @return [Integer] The number of packets waiting on the read queue
    def read_queue_size(self):
        return self._read_queue_size

    # @return [Integer] The number of packets waiting on the write queue
    def write_queue_size(self):
        return self._write_queue_size

    # Should be implemented by subclass to return human readable connection string
    # which will be placed in log messages when connecting and during connection failures
    def connection_string(self):
        return self.name

    # Connects the interface to its target(s). Must be implemented by a
    # subclass.
    def connect(self):
        for protocol in self.read_protocols + self.write_protocols:
            protocol.connect_reset()

        periodic_cmds = self.options.get("PERIODIC_CMD")
        if periodic_cmds:
            self.scheduler = schedule.Scheduler()

            for log_dont_log, period, cmd_string in periodic_cmds:
                upper_log_dont_log = log_dont_log.upper()
                period = float(period)
                self.scheduler.every(period).seconds.do(
                    self.run_periodic_cmd,
                    log_dont_log=upper_log_dont_log,
                    cmd_string=cmd_string,
                )

            self.cancel_scheduler_thread = False
            self.scheduler_thread = threading.Thread(target=self.scheduler_thread_body, daemon=True)
            self.scheduler_thread.start()

    # Called immediately after the interface is connected.
    # By default this method will run any commands specified by the CONNECT_CMD option
    def post_connect(self):
        connect_cmds = self.options.get("CONNECT_CMD")
        if connect_cmds:
            for log_dont_log, cmd_string in connect_cmds:
                if log_dont_log.upper() == "DONT_LOG":
                    cmd(cmd_string, log_message=False)
                else:
                    cmd(cmd_string)

    # Indicates if the interface is connected to its target(s) or not. Must be:
    # implemented by a subclass.
    def connected(self):
        raise RuntimeError("connected not defined by Interface")

    # Disconnects the interface from its target(s). Must be implemented by a
    # subclass.
    def disconnect(self):
        periodic_cmds = self.options.get("PERIODIC_CMD")
        if periodic_cmds and self.scheduler_thread:
            self.cancel_scheduler_thread = True

        for protocol in self.read_protocols + self.write_protocols:
            protocol.disconnect_reset()

    def read_interface(self):
        raise RuntimeError("read_interface not defined by Interface")

    def write_interface(self, data, extra=None):
        raise RuntimeError("write_interface not defined by Interface")

    # Retrieves the next packet from the interface.
    # self.return [Packet] Packet constructed from the data. Packet will be
    #   unidentified (None target and packet names)
    def read(self):
        if not self.connected():
            raise RuntimeError(f"Interface not connected for read {self.name}")
        if not self.read_allowed:
            raise RuntimeError(f"Interface not readable {self.name}")

        try:
            first = True
            while True:
                extra = None
                # Protocols may have cached data for a packet, so initially just inject a blank string
                # Otherwise we can hold off outputting other packets where all the data has already
                # been received
                if not first or len(self.read_protocols) <= 0:
                    # Read data for a packet
                    data, extra = self.read_interface()
                    if data is None:
                        Logger.info(f"{self.name}: read_interface requested disconnect")
                        return None
                else:
                    data = b""
                    first = False

                for protocol in self.read_protocols:
                    data, extra = protocol.read_data(data, extra)
                    if data == "DISCONNECT":
                        Logger.info(
                            f"{self.name}: Protocol {protocol.__class__.__name__} read_data requested disconnect"
                        )
                        return None
                    if data == "STOP":
                        break
                if data == "STOP":
                    continue

                packet = self.convert_data_to_packet(data, extra)

                # Potentially modify packet
                for protocol in self.read_protocols:
                    packet = protocol.read_packet(packet)
                    if packet == "DISCONNECT":
                        Logger.info(
                            f"{self.name}: Protocol {protocol.__class__.__name__} read_packet requested disconnect"
                        )
                        return None
                    if packet == "STOP":
                        break
                if packet == "STOP":
                    continue

                # Return packet
                self.read_count += 1
                if not packet:
                    Logger.warn(f"{self.name}: Interface unexpectedly requested disconnect")
                return packet
        except RuntimeError as error:
            Logger.error(f"{self.name}: Error reading from interface")
            self.disconnect()
            raise error

    # Method to send a packet on the interface.
    # self.param packet [Packet] The Packet to send out the interface
    def write(self, packet):
        if not self.connected():
            raise RuntimeError(f"Interface not connected for write {self.name}")
        if not self.write_allowed:
            raise RuntimeError(f"Interface not writable {self.name}")

        with self._write():
            self.write_count += 1

            # Potentially modify packet
            for protocol in self.write_protocols:
                packet = protocol.write_packet(packet)
                if packet == "DISCONNECT":
                    Logger.info(
                        f"{self.name}: Protocol {protocol.__class__.__name__} write_packet requested disconnect"
                    )
                    self.disconnect()
                    return
                if packet == "STOP":
                    return

            data, extra = self.convert_packet_to_data(packet)

            # Potentially modify packet data
            for protocol in self.write_protocols:
                data, extra = protocol.write_data(data, extra)
                if data == "DISCONNECT":
                    Logger.info(f"{self.name}: Protocol {protocol.__class__.__name__} write_data requested disconnect")
                    self.disconnect()
                    return
                if data == "STOP":
                    return

            # Actually write out data if not handled by protocol:
            self.write_interface(data, extra)

            # Potentially block and wait for response
            for protocol in self.write_protocols:
                packet, data, extra = protocol.post_write_interface(packet, data, extra)
                if packet == "DISCONNECT":
                    Logger.info(
                        f"{self.name}: Protocol {protocol.__class__.__name__} post_write_packet requested disconnect"
                    )
                    self.disconnect()
                    return
                if packet == "STOP":
                    return
        return None

    # Writes preformatted data onto the interface. Malformed data may cause
    # problems.
    # self.param data [String] The raw data to send out the interface
    def write_raw(self, data, extra=None):
        if not self.connected():
            raise RuntimeError(f"Interface not connected for write_raw {self.name}")
        if not self.write_raw_allowed:
            raise RuntimeError(f"Interface not raw writable {self.name}")

        with self._write():
            self.write_interface(data, extra)

    # Wrap all writes in a mutex and handle errors
    @contextmanager
    def _write(self):
        self.write_mutex.acquire()
        try:
            yield
        except WriteRejectError as error:
            Logger.error(f"{self.name}: Write rejected by interface {repr(error)}")
            raise error
        except RuntimeError as error:
            Logger.error(f"{self.name}: Error writing to interface")
            self.disconnect()
            raise error
        finally:
            self.write_mutex.release()

    def as_json(self):
        config = {}
        config["name"] = self.name
        config["state"] = self.state
        config["clients"] = self.num_clients()
        config["txsize"] = self.write_queue_size()
        config["rxsize"] = self.read_queue_size()
        config["txbytes"] = self.bytes_written
        config["rxbytes"] = self.bytes_read
        config["txcnt"] = self.write_count
        config["rxcnt"] = self.read_count
        return config

    # Start raw logging for this interface
    def start_raw_logging(self):
        if not self.stream_log_pair:
            self.stream_log_pair = StreamLogPair(self.name)
        self.stream_log_pair.start()

    # Stop raw logging for this interface
    def stop_raw_logging(self):
        if self.stream_log_pair:
            self.stream_log_pair.stop()

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, name):
        self.__name = name
        if self.stream_log_pair:
            self.stream_log_pair.name = name

    # Copy settings from this interface to another interface. All instance
    # variables are copied except for num_clients, read_queue_size,
    # and write_queue_size since these are all specific to the operation of the
    # interface rather than its instantiation.
    #
    # self.param other_interface [Interface] The other interface to copy to
    def copy_to(self, other_interface):
        other_interface.name = self.name
        other_interface.target_names = self.target_names[:]
        other_interface.cmd_target_names = self.cmd_target_names[:]
        other_interface.tlm_target_names = self.tlm_target_names[:]
        other_interface.connect_on_startup = self.connect_on_startup
        other_interface.auto_reconnect = self.auto_reconnect
        other_interface.reconnect_delay = self.reconnect_delay
        other_interface.disable_disconnect = self.disable_disconnect
        other_interface.packet_log_writer_pairs = self.packet_log_writer_pairs[:]
        other_interface.routers = self.routers[:]
        other_interface.cmd_routers = self.cmd_routers[:]
        other_interface.read_count = self.read_count
        other_interface.write_count = self.write_count
        other_interface.bytes_read = self.bytes_read
        other_interface.bytes_written = self.bytes_written
        if self.stream_log_pair:
            other_interface.stream_log_pair = self.stream_log_pair[:]
        # num_clients is per interface so don't copy
        # read_queue_size is the number of packets in the queue so don't copy
        # write_queue_size is the number of packets in the queue so don't copy
        for option_name, option_values in self.options.items():
            other_interface.set_option(option_name, option_values)
        other_interface.protocol_info = []
        for protocol_class, protocol_args, read_write in self.protocol_info:
            if not read_write == "PARAMS":
                other_interface.add_protocol(protocol_class, protocol_args, read_write)

    # Set an interface or router specific option
    # self.param option_name name of the option
    # self.param option_values array of option values
    def set_option(self, option_name, option_values):
        option_name_upcase = option_name.upper()

        # CONNECT_CMD and PERIODIC_CMD are special because there could be more than 1
        # so we store them in an array for processing during connect()
        if option_name_upcase == "PERIODIC_CMD" or option_name_upcase == "CONNECT_CMD":
            # OPTION PERIODIC_CMD LOG/DONT_LOG 1.0 "INST COLLECT with TYPE NORMAL"
            self.options[option_name_upcase] = self.options.get(option_name_upcase, [])
            self.options[option_name_upcase].append(option_values[:])
        else:
            self.options[option_name_upcase] = option_values[:]

    # Called to convert the read data into a OpenC3 Packet object
    #
    # self.param data [String] Raw packet data
    # self.return [Packet] OpenC3 Packet with buffer filled with data
    def convert_data_to_packet(self, data, extra):
        packet = Packet(None, None, "BIG_ENDIAN", None, data)
        packet.extra = extra
        return packet

    # Called to convert a packet into the data to send
    #
    # @param packet [Packet] Packet to extract data from
    # @return data
    def convert_packet_to_data(self, packet):
        return (
            packet.buffer,
            packet.extra,
        )  # Copy buffer so logged command isn't modified

    # Called to read data and manipulate it until enough data is
    # returned. The definition of 'enough data' changes depending on the
    # protocol used which is why this method exists. This method is also used
    # to perform operations on the data before it can be interpreted as packet
    # data such as decryption. After this method is called the post_read_data
    # method is called. Subclasses must implement this method.
    #
    # self.return [String] Raw packet data
    def read_interface_base(self, data, extra=None):
        self.read_raw_data_time = datetime.now(timezone.utc)
        self.read_raw_data = data
        self.bytes_read += len(data)
        if self.stream_log_pair:
            self.stream_log_pair.read_log.write(data)

    # Called to write data to the underlying interface. Subclasses must
    # implement this method and call super to count the raw bytes and allow raw
    # logging.
    #
    # self.param data [String] Raw packet data
    # self.return [String] The exact data written
    def write_interface_base(self, data, extra=None):
        self.written_raw_data_time = datetime.now(timezone.utc)
        self.written_raw_data = data
        self.bytes_written += len(data)
        if self.stream_log_pair:
            self.stream_log_pair.write_log.write(data)

    def add_protocol(self, protocol_class, protocol_args, read_write):
        protocol_args = protocol_args[:]
        protocol = protocol_class(*protocol_args)
        match read_write:
            case "READ":
                self.read_protocols.append(protocol)
            case "WRITE":
                self.write_protocols.insert(0, protocol)
            case "READ_WRITE" | "PARAMS":
                self.read_protocols.append(protocol)
                self.write_protocols.insert(0, protocol)
            case _:
                raise RuntimeError(
                    f"Unknown protocol descriptor {read_write}. Must be 'READ', 'WRITE', or 'READ_WRITE'."
                )
        self.protocol_info.append([protocol_class, protocol_args, read_write])
        protocol.interface = self
        return protocol

    def interface_cmd(self, cmd_name, *cmd_args):
        match cmd_name:
            case "clear_counters":
                self._write_queue_size = 0
                self._read_queue_size = 0
                self.bytes_written = 0
                self.bytes_read = 0
                self.write_count = 0
                self.read_count = 0

    def protocol_cmd(self, cmd_name, *cmd_args, read_write="READ_WRITE", index=-1):
        read_write = str(read_write).upper()
        if read_write not in ["READ", "WRITE", "READ_WRITE"]:
            raise RuntimeError(f"Unknown protocol descriptor {read_write}. Must be 'READ', 'WRITE', or 'READ_WRITE'.")
        handled = False

        if index >= 0 or read_write == "READ_WRITE":
            # Reconstruct full list of protocols in correct order
            protocols = []
            read_protocols = self.read_protocols
            write_protocols = self.write_protocols[:]
            write_protocols.reverse()
            read_index = 0
            write_index = 0
            for (
                _,
                _,
                protocol_read_write,
            ) in self.protocol_info:
                match protocol_read_write:
                    case "READ":
                        protocols.append(read_protocols[read_index])
                        read_index += 1
                    case "WRITE":
                        protocols.append(write_protocols[write_index])
                        write_index += 1
                    case "READ_WRITE" | "PARAMS":
                        protocols.append(read_protocols[read_index])
                        read_index += 1
                        write_index += 1

            for protocol_index, protocol in enumerate(protocols):
                result = None
                # If index is given that is all that matters
                if index == protocol_index or index == -1:
                    result = protocol.protocol_cmd(cmd_name, *cmd_args)
                if result:
                    handled = True
        elif read_write == "READ":  # and index == -1
            for protocol in self.read_protocols:
                result = protocol.protocol_cmd(cmd_name, *cmd_args)
                if result:
                    handled = True
        else:  # read_write == 'WRITE' and index == -1
            for protocol in self.write_protocols:
                result = protocol.protocol_cmd(cmd_name, *cmd_args)
                if result:
                    handled = True
        return handled

    def run_periodic_cmd(self, log_dont_log, cmd_string):
        if self.connected():
            try:
                if log_dont_log == "DONT_LOG":
                    cmd(cmd_string, log_message=False)
                else:
                    cmd(cmd_string)
            except Exception:
                Logger.error(f"Error sending periodic cmd({cmd_string}):\n{traceback.format_exc()}")

    def scheduler_thread_body(self):
        next_time = time.time()
        while True:
            if self.cancel_scheduler_thread:
                break

            self.scheduler.run_pending()

            if self.cancel_scheduler_thread:
                break
            next_time = next_time + 0.1  # Max 10 Hz
            sleep_time = next_time - time.time()
            if sleep_time > 0.1:
                time.sleep(0.1)
            elif sleep_time > 0:
                time.sleep(sleep_time)



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

WHITELIST = []

from .api_shared import *
from .cmd_api import *
from .config_api import *
from .interface_api import *
from .limits_api import *
from .offline_access_api import *
from .router_api import *
from .settings_api import *
from .stash_api import *
from .target_api import *
from .tlm_api import *
try:
    from openc3enterprise.api.cmd_authority_api import *
except ModuleNotFoundError:
    # ModuleNotFoundError expected in Open Source Edition
    pass
